#----------------------------------------------------------------------------\
# NetReg 2.0 (pre-release devel version)
#       DHCPD Registration Frontend
#
# Version 1.3: 
# Copyright (C) 1999-2002 Peter Valian, Southwestern University
# Version 2.0:
# Portions Copyright (C) 1999-2002 Peter Valian, Southwestern University.
# Portions Copyright (C) 2002 David Bussenschutt, Griffith University, Australia.
# Portions Copyright (C) 2002 Robert Lowe,  Lawrence University, Wisconsin.
# Portions Copyright (C) 2002-2003 Michael McMaster, Griffith University, Australia.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Version 1.3: http://www.netreg.org/
# Version 2.0: http://www.sourceforge.net/projects/netreg/
#
# Version 1.3 code contributions by:
#   Peter Valian <valianp@southwestern.edu>,  Southwestern University.
#     Author of Version 1.3rc2 
#   Bobby Clark <bobby.clark@eku.edu>, Eastern Kentucky University 
#     IMAP Authentication
#   Michael Grinnell <grinnell@american.edu>, American University
#     LDAP Authentication
#
# Version 2.0 code contributions by:
#   Robert Lowe <robert.h.lowe@lawrence.edu>, Lawrence University, Wisconsin.
#     CIDR (Subnet Masks etc) Handling/Notation
#     Alpha Testing, and bug chasing in David's Code.
#   David Bussenschutt <D.Bussenschutt@mailbox.gu.edu.au>, Griffith University, Queensland, Australia.
#     Much re-arrangement, and fixes, including but not limited to:
#     Perl 5 Syntax, and modules
#     File Locking
#     OMAPI Support
#     Remote dhcpdmanager daemon, and client support (functionally similar to omapi, but more fully implemented)
#     Improved Error Handling
#     Use of -w for improved code
#     LDAP group handling, for Permissions.
#     install_script
#   Michael McMaster <m4k3r@users.sourceforge.net>, Griffith University,  Queensland, Australia.
#     DBI/Oracle Support
#     HTML extraction/templating
#     Permissions handling (access controls)
#     Cookie based authentication
#     Interface for editing dhcp options/groups and hosts.
#     Synchronisation between hosts
#     Scalability improvements
#     Various code clean-ups
#     Consistency checker for host information 
#----------------------------------------------------------------------------/
package Admin;
use strict;
use warnings;
#----------------------------------------------------------------------------\\
#
# Exports
#
#----------------------------------------------------------------------------//
use Exporter();
our @ISA = qw(Exporter);

our @EXPORT = qw( );
our @EXPORT_OK = qw( server_manager find_conf find_lease subnet_overview
                     view_subnet display_popup expire_host display_history
                     get_color);

#----------------------------------------------------------------------------\\
#
# Imported Modules
#
#----------------------------------------------------------------------------//

# For timelocal (calculating expiration dates)
use Time::Local;
use Socket;

# Netreg modules.
use Subnets       qw( ip_in_subnet ip_in_subnet_urange read_domains
                      ip2dec dec2ip ip_in_subnet_d ip_in_subnet_urange_d);

use Variables     qw( variables_init $LDAP_URL $ADMINPATH $GFX $IP $FGREP $DEBUG
                      $DHCPDMASTERLOCK $SYNC_SCRIPT $DIG @ZONES $TIMEOFFSET $PS $GREP);

use Html          qw( get_input html_init admin_error get_cookies 
                      print_html_file);

use Authenticate  qw( user priv );

use Files_Remote  qw( read_leases_remote read_conf_new_remote 
                      delete_conf_new_remote status_dhcpd_remote 
                      read_ip_leases_remote);

use Database      qw( get_dbh db_connect db_get_unregip
                      db_del_host db_lock_host db_get_all db_active
                      get_all_subnets get_subnet_desc
                      db_get_static_ips_macs);

use NetregMisc    qw( open_with_lock make_time append_error_log );

#----------------------------------------------------------------------------\\
#
# Constants
#
#----------------------------------------------------------------------------//
# Optional mod_perl support.
# Sets a constant, and ensures the correct exit() function is called.
use constant IS_MODPERL => $ENV{MOD_PERL};
use subs qw(exit);
# Select the correct exit function
*exit = IS_MODPERL ? \&Apache::exit : sub { CORE::exit };


#----------------------------------------------------------------------------\\
#
# SQL init
#
#----------------------------------------------------------------------------//
# Prepare database queries. This function is only called by db_connect in
# Database.pm

my ($sql_count_match,$sql_match,$sql_all_registered,$sql_statics,$sql_log);
sub db_init {
    my $dbh = shift;

    $sql_log = $dbh->prepare (
            "select * from interface_log where mac = ? order by entrydate desc");

    $sql_statics = $dbh->prepare("select mac,cidr_subnet,cidr_mask from interface where ".
                                "status = 'S'");

    $sql_all_registered = $dbh->prepare("select mac,statusdate from interface where ".
                                "status = 'R'");

    $sql_count_match = $dbh->prepare ("select count(mac) from interface ".
           "where upper(mac) like ? or upper(hostname) like ? or upper(assetno) like ? or ".
           "upper(owner) like ? or upper(domain) like ? or upper(status) like ? or ".
           "upper(platform) like ? or upper(comments) like ? or ip like ? or upper(group_id) like ?");           

    $sql_match = $dbh->prepare ("select * from interface ".
           "where upper(mac) like ? or upper(hostname) like ? or upper(assetno) like ? or ".
           "upper(owner) like ? or upper(domain) like ? or upper(status) like ? or ".
           "upper(platform) like ? or upper(comments) like ? or ip like ? or upper(group_id) like ?");
}
#----------------------------------------------------------------------------\\
#
# Exported Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# display_popup: Displays a pop-up window for entry of data for expiration 
# of hosts.
#
# Parameters: None.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub display_popup {
    my %form = get_input();
    if (priv('lock')) {
        print_html_file("admin_unregister_host.html", 
                        "\\\$"."MAC", $form{'MAC'});
    } elsif (priv('expire')) {
        print_html_file("admin_expire_host.html", 
                        "\\\$"."MAC", $form{'MAC'});
    } else {
        admin_error ("Insufficient Privileges", "Privilege");
    }
    exit();
}

#----------------------------------------------------------------------------\\
# get_host_status: Returns a string signifying the current status of a known
# host.
#
# Parameters: The expiry date of the host (in YYYYMMDD format)
#             The status of the host as kept in the db. ('R','S' or 'U'). Will
#                 return "Unknown" if the status is invalid.
#
# Returns: Either "Expired", "Unknown", "Static","Enabled" or "Locked"
#----------------------------------------------------------------------------//
sub get_host_status {
    my $expiry = shift;
    my $status = shift;

    # Check whether the current status is expired or not
    $expiry =~ /(\d\d\d\d)(\d\d)(\d\d)/ if $expiry;
    my $expired = 0;
    $expired = time() > timelocal(0,0,0,$3, $2 - 1, $1 - 1900) if
                            ($3 && $2 && $1);

    my %status_list = ('S','Static','R','Enabled','U','Locked');
    my $ret = $expired ? 'Expired' : ($status ? $status_list{$status} : 'Unknown');
   
    return $ret;
}

#----------------------------------------------------------------------------\\
# make_user_link: Returns some HTML that contains a <a href... > to a user
# directory (if $LDAP_URL is defined in Variables.pm), or simply the users name.
#
# Parameters: a user name.
#
# Returns: Some html.
#----------------------------------------------------------------------------//
sub make_user_link {
    my $user = shift;

    my $ret = $user; # Default null link without LDAP.
    if ($LDAP_URL) {
        $ret = "<a href=$LDAP_URL target=blank>$user</a>";
        $ret =~ s/<USERNAME>/$user/;
    }
    return $ret;
}

#----------------------------------------------------------------------------\\
# status_highlight: Returns a colour to be used to highlight a given host
# depending on its status.
#
# Parameters: A status given by get_host_status
#
# Returns: A colour string that may be used in HTML.
#----------------------------------------------------------------------------//
sub status_highlight {
    my $status = shift;
    my %colors = ('Enabled','green','Locked','red','Expired','yellow','Static','blue');
    return $colors{$status} || 'red';
}

#----------------------------------------------------------------------------\\
# bold_words: Inserts a HTML bold tag (<b></b>) around items in a list that
# match a given pattern.
#
# Parameters: A substring to make bold.
#             A list of strings to process.
#
# Returns: The original list of strings given, with any occurances of the 
#          substring bolded.
#----------------------------------------------------------------------------//
sub bold_words {
    my $bold = shift;
    my @ret = ();
    foreach (@_) {
        s/($bold)/<b>$1<\/b>/ig;
        push @ret, $_;
    }
    return @ret;
}

#----------------------------------------------------------------------------\\
# find_conf: Searches for a registered host matching the given pattern
# via a substring search. If no query is given, then the find form is displayed.
# This function is essentially the "Search Hosts" feature of the website.
#
# Parameters: None.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub find_conf {
    my %form = get_input();
    find_form('FC') unless $form{'query'};  # Will exit.

    print_html_file("admin_header.html");

    # Search through the database.
    my $dbh = get_dbh();
    my $query = "%".uc($form{'query'})."%";

    $sql_count_match->execute($query,$query,$query,$query,$query,$query,$query,$query,$query,$query);
    my @row = $sql_count_match->fetchrow_array();
    $sql_count_match->finish();
    $sql_match->execute($query,$query,$query,$query,$query,$query,$query,$query,$query,$query); 
 
    print_html_file("admin_findconf_start.html", 
                    "\\\$"."QUERY", $form{'query'},
                    "\\\$"."COUNT", $row[0]);

    my $color = get_color();
    # Display each host
    while (my @row = $sql_match->fetchrow_array()) {
        $color = get_color($color);
        my $platform = $row[7] || 'Unknown';

        my $date = $row[6];
        $date =~ s/(....)(..)(..)/$3-$2-$1/;

        my $host = "$row[1].$row[4]";
        my $mac = $row[0];
 
        my $group = $row[12];
        $group =~ s/\:/\ \:/g; # enable the group to wrap when displaying.

        # Check whether the current status is expired or not
        my $status = get_host_status($row[6], $row[5]);

        my $userlink = make_user_link($row[3]);
        ($host,$mac,$platform,$date,$group) = bold_words($form{'query'},$host,$mac,$platform,$date,$group);
        my $highlight = status_highlight($status);
  
        my $exp = $row[5] eq 'S' ? '' : 'Expire';
        my $del = $row[5] eq 'S' ? 'Delete' : '';

        print_html_file("admin_findconf_row.html", "\\^MAC", $mac,
                        "\\^COLOR", $color, "\\^HIGHLIGHT", $highlight,
                        "\\^USERLINK", $userlink, "\\^LINKMAC", $row[0],
                        "\\^HOST", $host, "\\^PLATFORM", $platform,
                        "\\^STATUS", $status, "\\^DATE", $date,
                        "\\^GROUP",$group,"\\^LINKGROUP",$row[12],
                        "\\^EXPIRE",$exp,"\\^DELETE",$del);
    }
    $sql_match->finish();
    print "</table>";
    print_html_file("admin_footer.html");
}

#----------------------------------------------------------------------------\\
# expire_host: Attempts to either expire or lock out a given host. Removes
# The host from the remote dhcpd conf files immediately. Does not apply to
# static hosts.
#
# Parameters: None.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub expire_host {
    my %form = get_input();

    if (priv('lock') && $form{'DATE'}) {
        if ($form{'DATE'} !~ /(\d+).(\d+).(\d+)/) {
            admin_error("Invalid date format.","Input");
            print_html_error("admin_footer.html");
        }
        $form{'DATE'} =~ /(\d+).(\d+).(\d+)/;
        my $year = $3 > 2030 ? 2030 : $3;
        $form{'DATE'} ="$year$2$1";
        if (!db_lock_host($form{'MAC'}, $form{'DATE'}, user(),
                          $form{'comment'})) {
            admin_error ("unable to lock host");
            print_html_file("admin_footer.html");
        }
    } else {
        if (priv('expire')) {
            if (!db_del_host($form{'MAC'}, user(), $form{'comment'})) {
                admin_error ("Unable to expire host");
                print_html_file("admin_footer.html");
            } 
        } else {
            admin_error ("Insufficient Privilege", "Privilege");
            print_html_file("admin_footer.html");
        }
    }
        
    # remove this host from the DHCP conf files immediately.
    my $err = delete_conf_new_remote($form{'MAC'},undef);
    if (!$err) {
         admin_error ("Unable to remove host from DHCP servers.");
         print_html_file("admin_footer.html");
    }
    find_conf();
}

#----------------------------------------------------------------------------\\
# change_timezone: Changes the timezone of a timestamp in the dhcpd.leases
# time format.
#
# Parameters: A timestamp in dhcpd.leases time format.
#
# Returns: An adjusted timestamp.
#----------------------------------------------------------------------------//
sub change_timezone {
    my $timestamp = shift;

    $timestamp =~ /(\d+)\/(\d+)\/(\d+)\s+(\d+):(\d+):(\d+)/;

    my $gmt_time = timelocal($6, $5, $4, $3, $2 - 1, $1 - 1900);
    my ($s,$m,$h,$dy,$mo,$yr,$wd,undef,undef) = localtime($gmt_time + ($TIMEOFFSET));
    return sprintf ("%02d%02d%02d %02d:%02d:%02d",$yr+1900,++$mo,$dy,$h,$m,$s);
}

#----------------------------------------------------------------------------\\
# find_lease: Searches for a dhcp lease matching the given pattern
# via a substring search. If no query is given, then the find form is displayed.
#
# Parameters: None.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub find_lease {
    my %form = get_input();
    find_form('FL') unless $form{'query'};

    print_html_file("admin_header.html");

    # we must have data to query for in dhcpd.leases
    # will produce all leases for a host, MAC or IP (or timestamps)
    my $leases = read_leases_remote(undef,undef,$form{'query'}); 
    print_admin_footer() unless $leases;

    my $ltotal = $#{$leases} + 1;
    # Sort by timestart
    my @sorted_leases = sort ({(split (/#/,$b))[2] cmp (split (/#/,$a))[2];} @$leases);

    print_html_file("admin_findlease_start.html", "\\^QUERY", $form{'query'});

    my $cnt = my $acnt = my $color = 0;
    foreach my $lease (@sorted_leases) {
        my ($bindstate,$ip,$timestart,$timeend,$mac) = split /#/, $lease;
        $timestart= change_timezone($timestart);
        $timeend= change_timezone($timeend);

        $color = get_color($color);
        my @row = db_get_all($mac);

        # Check whether the current status is expired or not
        my $status = get_host_status($row[6], $row[5]);
        my $highlight = status_highlight($status);
        my $display_mac = $mac;
        ($mac,$ip,$timestart,$timeend) = bold_words($form{'query'},$mac,$ip,$timestart,$timeend);

        my $print_bindstate = "";
        if ($bindstate eq "active") {
            $print_bindstate = "<img src=$GFX/reddot.gif border=0>";
            $acnt++;
        }
        print_html_file("admin_findlease_row.html", "\\^MAC", $display_mac,
                        "\\^COLOR", $color, "\\^IP", $ip,
                        "\\^BINDSTATE", $print_bindstate,
                        "\\^TIMESTART", $timestart,
                        "\\^TIMEEND", $timeend, "\\^STATUS", $status,
                        "\\^HIGHLIGHT", $highlight);
        $cnt++;
    }

    # NOTE: This html is a hack. It shouldn't be here, it should be above
    # The table tag. It only works because most browsers are nice and reposition it
    # for us.
    print "<p>Found/Active: <b>$cnt</b> / <b>$acnt</b> matches\n";
    print "<br>Total Leases: <b>$ltotal</b><p>\n";
    print "</table>";
    print_html_file("admin_footer.html");
}

#----------------------------------------------------------------------------\\
# find_form: Displays a simple search form.
#
# Parameters: The type of form to display. May be "FL" (find lease) or
#             "FC" (find conf/registered mac).
#
# Returns: Nothing useful (will exit)
#----------------------------------------------------------------------------//
sub find_form {
    my $action = shift;

    print_html_file("admin_header.html");

    my $type = $action eq 'FL' ? 'Leases' : 'Registration Database';

    print_html_file("admin_find_form.html", "\\^TYPE", $type,
                    "\\^ACTION", $action);
    print_html_file("admin_footer.html");
}

#----------------------------------------------------------------------------\\
# server_manager: Displays a summary of current remote server status.
#
# Parameters: None.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub server_manager {
    append_error_log("Server manager starts here") if $DEBUG >=1;
    print_html_file("admin_header.html");
    print "<h2>Server Manager</h2>";
    print `uptime`."<p>";

    my %server = ("1:named"=> "DNS",
                  "2:httpd"=> "HTTP" );

    my (@list, $pid, $start, $status,$color);

    print_html_file("admin_servermanager_start.html");

    foreach my $proc (sort keys %server) {
        (undef,my $procname) = split /:/, $proc;
        my $pso = `$PS | $GREP $procname | $GREP -v grep`;
        $pso =~ s/\s+/ /g;
        @list = split / /, $pso;
        $pid = $list[1] || "";
        $start = $list[8] || "";
        if ($start =~ /^[a-z]+$/i) {
            $start .= " $list[9]";
        }

        ($status,$color) = $pid ? ("UP","green") : ("DOWN","red");

        print_html_file("admin_servermanager_row.html",
                        "\\^VAL1", $server{$proc}, "\\^VAL2", $status,
                        "\\^VAL3", $pid, "\\^VAL4", $start,
                        "\\^COLOUR", $color);
    }
    append_error_log("Server manager almost midway here...") if $DEBUG >=2;

    my $server = status_dhcpd_remote(undef);

    append_error_log("Server manager past midway here...") if $DEBUG >=2;
   
    foreach (@$server) {
        append_error_log("Server manager checked for $_ here") if $DEBUG >=2;
        my ($ip,$dhcpd_status,$pid,$info) = split /,/;

        my $packed_ip = inet_aton($ip);
        my $hostname = gethostbyaddr($packed_ip, AF_INET);

        # Check DNS also...
        my $dns_status = 0;
        my $command = "$DIG @"."$ip $ZONES[0] ANY";
        my $result = `$command`;
        $dns_status = 1 if ($result =~ /status\: NOERROR/);

        my ($status,$color) = ('UP','green');
        if ($dhcpd_status =~ /^UP/ && !$dns_status) {
            ($status,$color) = ('DNS down', 'red');
        } elsif ($dhcpd_status !~ /^UP/ && $dns_status) {
            ($status,$color) = ('DHCP down', 'red');
        } elsif ($dhcpd_status  !~ /^UP/ && !$dns_status) {
            ($status,$color) = ('DHCP/DNS down', 'red');
        }

        print_html_file("admin_servermanager_row.html",
                        "\\^VAL1", $hostname, "\\^VAL2", $status,
                        "\\^VAL3", $pid, "\\^VAL4", $info,
                        "\\^COLOUR", $color);

    }
    append_error_log("Server manager ends here") if $DEBUG >=1;
    print "</table></th></tr></table>\n";
    print_html_file("admin_footer.html");
}

#----------------------------------------------------------------------------\\
# view_subnet: Displays hosts currently in a given subnet
#
# Parameters: None.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub view_subnet {
    my %form = get_input();
    print_html_file("admin_header.html");

    if ($form{'subnet'} !~ /^[\d\.\/]/) {
        print_html_file("Invalid subnet supplied.");
        print_html_file("admin_footer.html");
    }
    print_html_file("admin_subnet_start.html", "\\^SUBNET", $form{'subnet'}); 

    my %lease_ips = ();
    my $leases = read_leases_remote(undef,$form{'subnet'});
    foreach (@$leases) {
        /lease ([\d\.]+)/;
        my $ip = $1;
        /hardware ethernet ([\:a-f0-9]+)/si;
        $lease_ips{$ip} = $1 if $1;
    }

    # Get a list of statics for this subnet
    my ($subnet, $mask) = split('/', $form{'subnet'});
    my $static_ips = db_get_static_ips_macs($subnet,$mask);
    %lease_ips= (%lease_ips, %$static_ips);

    my @ip_sorted = sort ({ip2dec($a) <=> ip2dec($b)} (keys %lease_ips));

    my $color;
    foreach (@ip_sorted) {
        my $ip = $_;
        my $mac = $lease_ips{$_};
        $color = get_color($color);
        my @row = db_get_all($mac);
      
        my $host = $row[1] ? "$row[1].$row[4]" : 'Unknown';
        my $platform = $row[7] || 'Unknown';
        my $date = $row[6] || 'Unregistered';

        # Check whether the current status is expired or not
        my $status = get_host_status($row[6], $row[5]);

        my $exp = $row[5] && $row[5] eq 'S' ? '' : 'Expire';
        my $del = $row[5] && $row[5] eq 'S' ? 'Delete' : '';
        my $upd = $row[0] ? 'Update' : '';

        my $group = $row[12];
        $group =~ s/\:/\ \:/g;  #Enable line wrapping.

        my $highlight = status_highlight($status);
        $date =~ s/(....)(..)(..)/$3-$2-$1/ unless $date eq 'Unregistered';
        print_html_file("admin_subnet_row.html", "\\^MAC", $mac,
                        "\\^COLOR", $color, "\\^LINKMAC", $mac,
                        "\\^HOST", $host, "\\^PLATFORM", $platform,
                        "\\^DATE", $date, "\\^STATUS", $status,
                        "\\^HIGHLIGHT", $highlight, "\\^IP", $ip,
                        "\\^GROUP",$group,"\\^LINKGROUP",$row[12],
                        "\\^EXPIRE",$exp,"\\^DELETE",$del,"\\^UPDATE",$upd);
    }
    print "</table>";
    print_html_file("admin_footer.html");
}

#----------------------------------------------------------------------------\\
# subnet_overview: Displays the number hosts that are currently active per
# subnet. This method has been heavily optimized for speed, because thousands
# of hosts have the chance to take a very long time.
#
# Parameters: None.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub subnet_overview {
    print_html_file("admin_header.html");

    # Sort the subnets based on decimal IP, not alphanumeric.
    my @temp_subnets = get_all_subnets();
    my @subnets;
    foreach (@temp_subnets) {
        push @subnets, ip2dec($_)."#$_";
    }
    @subnets = sort ({(split (/#/,$a))[0] <=> (split (/#/,$b))[0]} @subnets);

    my $atotal = my $utotal = my $total = my $stotal = 0;
    my (%reg_ip, %ureg_ip, %sreg_ip);

    my @lease_ips = ();
    # NOTE: This is in a do() loop to ensure the potentially
    # large $leases var goes out of scope quickly
    # and the ram is recovered.
    do {
        my $leases = read_ip_leases_remote();
        if (!$leases) {
            admin_error ("Could not get lease files.", "Network");
            print_html_file("admin_footer.html");
        }
        @lease_ips = sort ({ip2dec($a) <=> ip2dec($b)} @$leases);
    };

    # in the interest of efficiency, each ip is assumed to be in the 
    # subnet of the previous ip.
    # if this is incorrect, then do a complete search starting at the
    # last found subnet.
    my ($dec_subnet,$last_subnet,$last_subnet_mask) = split('[#/]', $subnets[0]);# initialise the last found subnet with 
    my $last_subnet_index = 0;
    foreach my $lease_IP (@lease_ips) {
        my $dec_ip = ip2dec($lease_IP);
        # Test the last known subnet.
        if (ip_in_subnet_d($dec_ip, $dec_subnet, $last_subnet_mask)) {
           if (ip_in_subnet_urange_d($dec_ip,$last_subnet, $last_subnet_mask)) {
                $ureg_ip{"$last_subnet/$last_subnet_mask"}++;
                $utotal++;
            } else {
                $reg_ip{"$last_subnet/$last_subnet_mask"}++;
                $atotal++;
            }
            next;
        }

        # If we get to here, then we need to do a complete search of all subnets.
        # But, since the subnet list is ALSO sorted, we can speed things up by starting
        # from the NEXT subnet in the list.
        my $emergency_stop_index = $last_subnet_index; # should NOT continue past this point.

        while (1) { # Loop until exited with "last"
            $last_subnet_index++;
            if ($last_subnet_index > $#subnets) {
                $last_subnet_index = 0;
            }
            if ($last_subnet_index == $emergency_stop_index) {
                # grr!! this shouldn't happen.
                print STDERR "Could not find subnet for IP $lease_IP!!\n";
                last;
            }
    
            my ($dec,$subnet,$mask) = split('[#/]',$subnets[$last_subnet_index]);
            if (ip_in_subnet_d($dec_ip, $dec, $mask)) {
                if (ip_in_subnet_urange_d($dec_ip,$subnet,$mask)) {
                    $ureg_ip{"$subnet/$mask"}++;
                    $utotal++;
                } else {
                    $reg_ip{"$subnet/$mask"}++;
                    $atotal++;
                }
                $last_subnet = $subnet;
                $last_subnet_mask = $mask;
                $dec_subnet = $dec;
                last;
            }
        }
    }

    # Machines are no longer registered for a given subnet, but are registered 
    # for all subnets. Find out how many active registrations there are.
    my $dbh = get_dbh();
    $sql_all_registered->execute();
    my $time = make_time();
    $time =~ s/^(\d+).*$/$1/;
    while (my @row = $sql_all_registered->fetchrow_array()) {
        $total++ if ($row[1] > $time);
    }
    $sql_all_registered->finish();

    # Find out how many static hosts exist.
    $sql_statics->execute();
    while (my @row = $sql_statics->fetchrow_array()) {
        $stotal++;
        $sreg_ip{"$row[1]/$row[2]"}++;
    }
    $sql_statics->finish();

    $utotal = '&nbsp;0' unless $utotal;
    $atotal = '&nbsp;0' unless $atotal;
    $total = '&nbsp;0' unless $total;
    $stotal = '&nbsp;0' unless $stotal;
    print_html_file("admin_subnetoverview_start.html", "\\^UTOTAL", $utotal,
                    "\\^ATOTAL", $atotal, "\\^TOTAL", $total, 
                    "\\^STOTAL", $stotal);

    foreach my $SUBNET (@subnets) {
        $SUBNET =~ s/^.*#//;
        print "<tr><td align=right>";
        print "<a href=$ADMINPATH/admin.cgi?action=VS&subnet=$SUBNET>";
        print "$SUBNET</a></td>";
        print "<td width=350 bgcolor=\#eeeeee>";
        my $greenwidth = $reg_ip{$SUBNET} || 0;
        my $redwidth = $ureg_ip{$SUBNET} || 0;
        my $bluewidth = $sreg_ip{$SUBNET} || 0;
        my $s_cnt = $bluewidth;
        $bluewidth *= 2;
        $greenwidth *= 2;
        $redwidth *= 10;
        my $a_cnt = $reg_ip{$SUBNET} || 0;
        my $u_cnt = $ureg_ip{$SUBNET} || 0;

        print "<table width=$greenwidth bgcolor=green border=0 hspace=0 cellpadding=0 cellspace=0>".
              "<tr><td align=center><font color=white size=1><b>$a_cnt</b></font></td></tr></table>" if $greenwidth;
        print "<table width=$redwidth bgcolor=red border=0 hspace=0 cellpadding=0 cellspace=0>".
              "<tr><td align=center><font color=white size=1><b>$u_cnt</b></font></td></tr></table>" if $redwidth;
        print "<table width=$bluewidth bgcolor=blue border=0 hspace=0 cellpadding=0 cellspace=0>".
              "<tr><td align=center><font color=white size=1><b>$s_cnt</b></font></td></tr></table>" if $bluewidth;
        
        print "</td><td><a ";
        my ($sub,$mask) = split('/',$SUBNET);
        my $current_desc = get_subnet_desc($sub,$mask);
        print "href=$ADMINPATH/admin.cgi?action=VS&subnet=$SUBNET>",$current_desc,"</a>";
        print "</td></tr>\n";
    }
    print "</table></form>";
    print_html_file("admin_footer.html");
}

#----------------------------------------------------------------------------\\
# display_history: Displays the history (in the log files) of the given MAC.
#
# Parameters: None.
#
# Returns: Nothing.
#----------------------------------------------------------------------------//
sub display_history {
    my %form = get_input();
    find_form('FC') unless $form{'MAC'};
    
    my $mac = $form{'MAC'};

    # Determine the vendor of this MAC address.
    $mac =~ /^(..).(..).(..)/;
    my $oui = uc ("$1-$2-$3");
    my $vendor = `$FGREP $oui %%%paths{'cgi-bin'}%%%oui.txt`;
    $vendor =~ s/$oui\s*//;

    print_html_file("admin_header.html");
    my $dbh = get_dbh();
    print_html_file("admin_displayhistory_start.html", "\\^MAC", $mac,"\\^TYPE",$vendor);

    $sql_log->execute(uc($mac));

    my $color = get_color();
    while (my @row = $sql_log->fetchrow_array) {
        # support multi-line comments. replace \n with <br />
        $row[11] =~ s/\n/\<br \/\>/go if $row[11];

        my $status = get_host_status($row[6], $row[5]);
        my $userlink = make_user_link($row[3]);

        # Reformat the dates.
        $row[6] =~ s/(....)(..)(..)/$3-$2-$1/;
        $row[7] =~ s/(....)(..)(..)/$3-$2-$1/;
    
        print_html_file("admin_displayhistory_row.html", "\\^COLOR", $color,
                        "\\^HOST", $row[1], "\\^ASSET", $row[2], "\\^OWNER", $userlink,
                        "\\^DOMAIN", $row[4], "\\^STATUS", $status, "\\^TIME", $row[6],
                        "\\^ENTRY", $row[7], "\\^CSO", $row[8], "\\^BROWSER", $row[9],
                        "\\^IP", $row[10], "\\^COMMENT", $row[11]);
        $color = get_color($color);
    }
    $sql_log->finish();
    print "</table>";
    print_html_file("admin_footer.html");
}

#----------------------------------------------------------------------------\\
# get_color: Returns a toggled colour for use in table rows.
#
# Parameters: The last colour returned by the method.
#
# Returns: An alternative colour.
#----------------------------------------------------------------------------//
sub get_color {
    my $color =  shift || ''; 
    return $color eq 'eeeeee' ? 'cccccc' : 'eeeeee';
}

1;
