#----------------------------------------------------------------------------\
# NetReg 2.0 (pre-release devel version)
#       DHCPD Registration Frontend
#
# Version 1.3: 
# Copyright (C) 1999-2002 Peter Valian, Southwestern University
# Version 2.0:
# Portions Copyright (C) 1999-2002 Peter Valian, Southwestern University.
# Portions Copyright (C) 2002 David Bussenschutt, Griffith University, Australia.
# Portions Copyright (C) 2002 Robert Lowe,  Lawrence University, Wisconsin.
# Portions Copyright (C) 2002-2003 Michael McMaster, Griffith University, Australia.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Version 1.3: http://www.netreg.org/
# Version 2.0: http://www.sourceforge.net/projects/netreg/
#
# Version 1.3 code contributions by:
#   Peter Valian <valianp@southwestern.edu>,  Southwestern University.
#     Author of Version 1.3rc2 
#   Bobby Clark <bobby.clark@eku.edu>, Eastern Kentucky University 
#     IMAP Authentication
#   Michael Grinnell <grinnell@american.edu>, American University
#     LDAP Authentication
#
# Version 2.0 code contributions by:
#   Robert Lowe <robert.h.lowe@lawrence.edu>, Lawrence University, Wisconsin.
#     CIDR (Subnet Masks etc) Handling/Notation
#     Alpha Testing, and bug chasing in David's Code.
#   David Bussenschutt <D.Bussenschutt@mailbox.gu.edu.au>, Griffith University, Queensland, Australia.
#     Much re-arrangement, and fixes, including but not limited to:
#     Perl 5 Syntax, and modules
#     File Locking
#     OMAPI Support
#     Remote dhcpdmanager daemon, and client support (functionally similar to omapi, but more fully implemented)
#     Improved Error Handling
#     Use of -w for improved code
#     LDAP group handling, for Permissions.
#     install_script
#   Michael McMaster <m4k3r@users.sourceforge.net>, Griffith University,  Queensland, Australia.
#     DBI/Oracle Support
#     HTML extraction/templating
#     Permissions handling (access controls)
#     Cookie based authentication
#     Interface for editing dhcp options/groups and hosts.
#     Synchronisation between hosts
#     Scalability improvements
#     Various code clean-ups
#     Consistency checker for host information 
#----------------------------------------------------------------------------/
package Database;
use strict;
use warnings;
#----------------------------------------------------------------------------\\
#
# Exports
#
#----------------------------------------------------------------------------//
use Exporter();
our @ISA = qw (Exporter);

our @EXPORT = qw( );
our @EXPORT_OK =  qw( get_dbh db_connect db_disconnect db_new_host db_del_host
                      db_lock_host db_already_registered db_next_serial
                      db_mac_locked 
                      db_get_unregip db_get_all db_active
                      db_unique_host db_del_static 
                      get_all_subnets get_subnet_desc db_get_static_ips_macs 
                      get_group_list db_get_mac db_get_campuses );

#----------------------------------------------------------------------------\\
#
# Imported Modules
#
#----------------------------------------------------------------------------//
# If using mod_perl, the $FindBin::RealBin directory will be set ONCE only
# see startup.pl
use FindBin;
use lib "$FindBin::RealBin"; # locate library path for below modules.

use Socket; # needed for inet_ntoa

# Use Perl Database support 
use DBI;

# Netreg modules.
use Variables qw( $DB_CONNECT $DB_USER $DB_PASS $ORACLE_HOME $IP);
use NetregMisc qw( append_error_log make_time);
use Html qw( admin_error );
use Authenticate qw (user );

#----------------------------------------------------------------------------\\
#
# Constants
#
#----------------------------------------------------------------------------//
# Optional mod_perl support.
# Sets a constant, and ensures the correct exit() function is called.
use constant IS_MODPERL => $ENV{MOD_PERL};
use subs qw(exit);
# Select the correct exit function
*exit = IS_MODPERL ? \&Apache::exit : sub { CORE::exit };

#----------------------------------------------------------------------------\\
#
# Global Variables
#
# These are shared within this perl module only.
# These variables are specifically shared between executions if using mod_perl
#----------------------------------------------------------------------------//
# A database handle. Shared between each execution.
my $dbh;

# The time at which the database handle ($dbh) was connected.
my $last_connect_time;

# Stored SQL queries.
my ($sql_log, $sql_check_existing_mac, $sql_update_mac, $sql_insert_mac,
    $sql_expire_mac, $sql_lock_mac, $sql_registered,
    $sql_locked, $sql_unregip, $sql_getall,$sql_get_mac,
    $sql_active, $sql_unique_host, $sql_del_static, $sql_group_list,
    $sql_all_subnets, $sql_subnet_desc, $sql_static_all, $sql_get_campuses );

#----------------------------------------------------------------------------\\
#
# Automatically Executed Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# END: Disconnect from the database and reclaim memory. SQL statements
# will automagically release their memory also.
#
# Returns: nothing useful.
#----------------------------------------------------------------------------//
END {
     $dbh->disconnect() if $dbh;
}

#----------------------------------------------------------------------------\\
#
# Exported Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# db_connect: Connects to the database (using info in Variables.pm), unless
# a connection already exists.
#
# Parameters: True if sql statements should not be prepared for use.
#
# Returns: the database handle, or undef on error.
#----------------------------------------------------------------------------//
sub db_connect {
    my $dont_init = shift;
    # Do not bother attempting to connect more than once every
    # 5 minutes
    if ($last_connect_time && (time() - $last_connect_time < 300) && $dbh) {
        return $dbh;
    } 
 
    # Check if we can simply return the current handle.
    # This is in an eval block because the ping call may do bad things
    # if the database connection is in fact dead.
    if ($dbh) {
       eval {
           if ($dbh->ping) {
               $last_connect_time = time(); 
               return $dbh;
           }
       }
    }
    
    # Either we've never connected before, or the database didn't ping.
    # Setting the environment could probably be somewhere else, but it isn't
    # hurting much being here (even if we're not using oracle).
    $ENV{'ORACLE_HOME'} = $ORACLE_HOME;
    $dbh = DBI->connect($DB_CONNECT, $DB_USER, $DB_PASS,
                        { AutoCommit => 1,
                          RaiseError => 0});
    $last_connect_time = time() if $dbh;
    prepare_sql() if $dbh;
    if ($dbh && !$dont_init) {
      &Authenticate::db_init($dbh);
      &Subnets::db_init($dbh);
      &Files_Remote::db_init($dbh);
      &Admin::db_init($dbh);
      
    } elsif (!$dbh) {
      print STDERR "Could not connect to database $DBI::errstr!!\n";
      append_error_log ("Could not connect to database !!\n");
      return undef;
    }
    return $dbh;
}

#----------------------------------------------------------------------------\\
# get_db: Returns a database handle. No checking is done to ensure the handle
# is valid.
#
# Parameters: None.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub get_dbh {
    return $dbh;
}

#----------------------------------------------------------------------------\\
# db_disconnect: Pretend to disconnect from the database. This function does
# not actually do anything because it's better (performance wise) to keep a
# permanent connection.
#
# Parameters: None
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub db_disconnect {
#    $dbh->disconnect() if defined $dbh;
#    undef $dbh;
}

#----------------------------------------------------------------------------\\
# db_del_static: Permamently delete a static host from the database
#
# Parameters: None
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub db_del_static {
    my $mac = shift;

    my $time = make_time();
    $time =~ s/^(\d+).*$/$1/;

    db_log($mac,undef,undef,undef,undef,'S', "$time", user(), undef,undef,
           "Static Host Deleted");
    $sql_del_static->execute(uc $mac);
    $sql_del_static->finish();
}

#----------------------------------------------------------------------------\\
# db_get_static_ips_macs: Creates a hash of static IP's to the mac addresses
# using those IP's.
#
# Parameters: The subnet containing the IP's
#             A subnet mask.
#
# Returns: A hash reference.
#----------------------------------------------------------------------------//
sub db_get_static_ips_macs {
    my $subnet = shift;
    my $mask = shift;
    my %ips;
    $sql_static_all->execute($subnet, $mask);
    while (my @row = $sql_static_all->fetchrow_array) {
        $ips{$row[9]} = $row[0];
    }
    $sql_static_all->finish();

    return \%ips;
}

#----------------------------------------------------------------------------\\
# db_log: Log the current status of a host (eg. when the status has just 
# changed) to the database.
#
# Parameters: mac, hostname, assetno, owner, domain, status ('R','U','S'),
#             statusdate, user, browser, ip, comment.
#             All fields except mac may be null.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub db_log {
    my ($mac, $hostname, $assetno, $owner, $domain, $status, $statusdate,
        $user, $browser, $unregip, $comment) = @_;

    my $entrydate = make_time();
  
    $user = 'Unknown' unless $user;
    $browser = $ENV{'HTTP_USER_AGENT'} unless $browser;
    $unregip = $IP unless $IP;

    my @row = db_get_all($mac);
    $hostname = $row[1] unless $hostname;
    $assetno = $row[2] unless $assetno;
    $assetno = -1 if ($assetno =~ /[a-z]/);
    $owner = $row[3] unless $owner;
    $domain = $row[4] unless $domain;
    $status = $row[5] unless $status;
    $statusdate = $row[6] unless $statusdate;   

    $sql_log->execute(uc $mac, $hostname, $assetno, $owner, $domain,
                      $status, $statusdate, $entrydate,$user, $browser,
                      $unregip, $comment);
    $sql_log->finish();
}

#----------------------------------------------------------------------------\\
# db_new_host: Create/update a host under ip management. 
#
# Parameters: mac, hostname, assetno, owner, domain, status
#             statusdate, user, browser, unregistered ip, comment,
#             subnet, mask, group,
#             True if updating an existing entry.
#
# Returns: 1 if successful, undef if unsuccessful.
#----------------------------------------------------------------------------//
sub db_new_host {
    my ($mac, $hostname, $assetno, $owner, $domain,
        $status, $statusdate, $user, $browser, $unregip, $comment,
        $subnet,$mask,$group,$isUpdate) = @_;

    # Oracle and Postgres convert empty/undef to NULL differently in database.
    for ( $DB_CONNECT ) {
	/DBI:Oracle/i	and do {   $subnet='' unless $subnet; $mask='' unless $mask;  last; };
	/DBI:Pg/i	and do {  $subnet=undef unless $subnet; $mask=undef unless $mask;  last; };
	die "Unknown Database connection string in \$DB_CONNECT: \"$DB_CONNECT\"  - exiting!\n";
    }

    $group = 'Roaming:Default' unless $group;
        
    # This should NOT happen, since it should already be checked
    # Before using this function.
    if (!$isUpdate && db_already_registered($mac) && $status ne 'S') {
      admin_error("MAC is already registered and active.\n");
      return undef;
    }
  
    if (db_mac_locked($mac)) {
      admin_error("MAC is currently 'unregistered'. Please see a superuser ".
                  "of the system.\n");
      return undef;
    }

    db_log ($mac, $hostname, $assetno, $owner, $domain, $status, $statusdate,
            $user, $browser, $unregip, $comment); 

    # The mac (primary key) may already be in the database
    $sql_check_existing_mac->execute(uc $mac);
    my @existing_row = $sql_check_existing_mac->fetchrow_array;
    $sql_check_existing_mac->finish();

    if (defined $existing_row[0]) {
        # Mac is already in database.
        $sql_update_mac->execute($hostname, $assetno, $owner,
                                 $domain, $status,$statusdate,$browser,
                                 $comment,$unregip,$subnet,$mask,$group,
                                 uc $mac) || return undef;
        $sql_update_mac->finish();
    } else { 
        $sql_insert_mac->execute(uc $mac, $hostname, $assetno, $owner, $domain,
                                 $status,$statusdate,$browser,$comment,
                                 $unregip,$subnet,$mask,$group) || return undef;
        $sql_insert_mac->finish();
    }
    return "1";
}

#----------------------------------------------------------------------------\\
# db_del_host: Expires the registration of a host uner IP management.
#              This may be performed by any user. Status of the host will remain
#              'R'
#
# Parameters: mac, user, comment
#
# Returns: 1 if successful, undef if unsuccessful.
#----------------------------------------------------------------------------//
sub db_del_host { 
    my ($mac, $user, $comment) = @_;

    # Do not touch this host if it is "unregistered"
    return undef if db_mac_locked($mac);
 
    # Log this action.
    my $time = make_time();
    $time =~ s/^(\d+).*$/$1/;
    my @mac_info = db_get_all($mac);
    db_log($mac,$mac_info[1],$mac_info[2],$mac_info[3],$mac_info[4],
           'R',"$time", $user,undef,undef, $comment);
         
    # Expire the host.
    $sql_expire_mac->execute($time,uc $mac) || return undef;
    $sql_expire_mac->finish();
    return "1";
}

#----------------------------------------------------------------------------\\
# db_lock_host: Locks out a given host under IP management until the given date.
# This may only be performed by superusers.
#
# Parameters: mac, date, user, comment
#
# Returns: 1 if successful, undef if unsuccessful.
#----------------------------------------------------------------------------//
sub db_lock_host {
    my $mac = shift;
    my $date = shift;
    my $user = shift;
    my $comment = shift;

    my @mac_info = db_get_all($mac);
    db_log($mac,$mac_info[1],$mac_info[2],$mac_info[3],$mac_info[4],
           'U',$date, $user, undef, undef, $comment);
    $sql_lock_mac->execute($date, uc $mac) || return undef;
    $sql_lock_mac->finish();
    return "1";
}

#----------------------------------------------------------------------------\\
# db_already_registered: Returns true if the MAC given is under IP Management,
# and the statusdate is sometime in the future. 
#
# Parameters: mac
#
# Returns: True if the MAC is registered and in date.
#----------------------------------------------------------------------------//
sub db_already_registered {
    my $mac = shift;

    my $time = make_time();
    $time =~ s/^(\d+).*$/$1/;

    $sql_registered->execute(uc $mac);
    my @rows = $sql_registered->fetchrow_array();
    $sql_registered->finish();
    return ($rows[1] && $rows[1] > $time ? 1 : 0);
}

#----------------------------------------------------------------------------\\
# db_mac_locked: Returns true if the MAC is currently locked out. ie,
# status of 'U' and not expired. 
#
# Parameters: mac
#
# Returns: True if the MAC is locked.
#----------------------------------------------------------------------------//
sub db_mac_locked {
    my $mac = shift;

    my $time = make_time();
    $time =~ s/^(\d+).*$/$1/;

    $sql_locked->execute(uc $mac);
    my @row = $sql_locked->fetchrow_array();
    $sql_locked->finish();

    return ($row[1] && $row[1] > $time ? 1 : 0);
}

#----------------------------------------------------------------------------\\
# db_next_serial: Returns the next serial for the given asset number.
# Serial numbers need to be kept unique.
#
# Parameters: asset number
#             the mac address using the new serial   
#
# Returns: The next serial number for the given asset number.
#----------------------------------------------------------------------------//
sub db_next_serial {
    my $asset = shift;
    my $mac = shift;

    # assetno in the form of asset-serial
    # (or asset if serial is 0)
    # where serial = 0,1,2,3 depending on no. of network cards.
    # Generally, there'll only ever be 1 network card anyway.
    my $sql = $dbh->prepare("select hostname,mac from interface where ".
                            "hostname like 'pc$asset%'");
    $sql->execute();

    my $serial = 0;
    while (my @row = $sql->fetchrow_array()) {
       if ($mac && $row[1] eq $mac) {
           # Allowed if a machine is de-registered.
           $row[0] =~ /pc\d+-(\d+)/;
           $serial = $1 || '0';
       } elsif ($row[0] =~ /pc\d+-(\d+)/ && $1 >= $serial) {
           $serial = $1 + 1;
       } elsif ($serial == 0) {
           $serial = 1;
       }
    }
    $sql->finish();
    return $serial;
}

#----------------------------------------------------------------------------\\
# db_get_unregip: Returns the unregistered IP of a given machine
#
# Parameters: MAC
#
# Returns: unregistered IP 
#----------------------------------------------------------------------------//
sub db_get_unregip {
    my $mac = shift;

    $sql_unregip->execute(uc $mac);
    my @row = $sql_unregip->fetchrow_array();
    $sql_unregip->finish();

    return $row[0];
}

#----------------------------------------------------------------------------\\
# db_get_all: Returns the all info for a given machine
#
# Parameters: MAC
#
# Returns: (mac,hostname,assetno,owner,domain,status,statusdate,...)
#----------------------------------------------------------------------------//
sub db_get_all {
    my $mac = shift;
 
    $sql_getall->execute(uc $mac);
    my @ret = $sql_getall->fetchrow_array(); 
    $sql_getall->finish();

    # reformat asset numbers
    $ret[2] = $ret[2] && $ret[2] < 0 ? 'Non-GU Asset' : sprintf("%06u",$ret[2] || 0);
    return @ret;
}

#----------------------------------------------------------------------------\\
# db_active: Returns true if the given MAC is under IP management, 'registered',
# AND not expired.
#
# Parameters: MAC
#
# Returns: true if the mac is currently active.
#----------------------------------------------------------------------------//
sub db_active {
    my $mac = shift;

    my $time = make_time();
    $time =~ s/^(\d+).*$/$1/; 

    $sql_active->execute(uc $mac);
    my @ret = $sql_active->fetchrow_array();
    $sql_active->finish();
    return ($ret[1] && $ret[1] > $time ? 1 : 0);
}

#----------------------------------------------------------------------------\\
# db_unique_host: Returns true if the given hostname is unique. Checks the
# database and DNS records.
#
# Parameters: hostname, domain, ip
#
# Returns: true if the host is unique.
#----------------------------------------------------------------------------//
sub db_unique_host {
    my $host = shift;
    my $domain = shift;
    my $ip = shift;

    my $existing_ip = gethostbyname("$host.$domain");
    return undef if $existing_ip && (inet_ntoa($existing_ip) ne $ip);
    
    $sql_unique_host->execute($host, $domain);

    my $time = make_time();
    $time =~ s/^(\d+).*$/$1/;

    while (my @ret = $sql_unique_host->fetchrow_array()) {
        do { $sql_unique_host->finish(); return 0} if $ret[1] > $time;
    }
    $sql_unique_host->finish();
    return 1;
}   

#----------------------------------------------------------------------------\\
# get_group_list: Lists mac addresses belonging to a given group.
#
# Parameters: Group id
#
# Returns: An array of mac addresses in that group.
#----------------------------------------------------------------------------//
sub get_group_list {
    my $group = shift;

    my @ret = ();
    $sql_group_list->execute($group);
    while (my @row = $sql_group_list->fetchrow_array) {
        push @ret, $row[0];
    }
    $sql_group_list->finish();
    return @ret;
}

#----------------------------------------------------------------------------\\
# get_subnet_desc: Returns a description for a given subnet
#
# Parameters: A subnet
#             subnet mask
#
# Returns: A description of the subnet.
#----------------------------------------------------------------------------//
sub get_subnet_desc {
    my $subnet = shift;
    my $mask = shift;

    $sql_subnet_desc->execute($subnet, $mask);
    my @row = $sql_subnet_desc->fetchrow_array();
    $sql_subnet_desc->finish;
    return $row[3];
}

#----------------------------------------------------------------------------\\
# db_get_campuses: returns a list of all valid campuses in the database
#
# Parameters: none.
#
# Returns: an array of campus names
#----------------------------------------------------------------------------//
sub db_get_campuses {

  $sql_get_campuses->execute();
  my @campuses = $sql_get_campuses->fetchrow_array();
  $sql_get_campuses->finish;

  return @campuses;
}

#----------------------------------------------------------------------------\\
# get_all_subnets: Returns a list of known subnets.
#
# Parameters: None.
#
# Returns: An array of "subnet/mask" type strings.
#----------------------------------------------------------------------------//
sub get_all_subnets {
    $sql_all_subnets->execute();
    my @ret = ();
    while (my @row = $sql_all_subnets->fetchrow_array) {
        push @ret, "$row[0]/$row[1]";
    }
    $sql_all_subnets->finish();
    return @ret;
}

#----------------------------------------------------------------------------\\
# db_get_mac: Returns the mac address of a given host.
# database handle.
#
# Parameters: Hostname, domain.
#
# Returns: A mac address.
#----------------------------------------------------------------------------//
sub db_get_mac {
    my ($hostname,$domain) = @_;
    $sql_get_mac->execute($hostname, $domain);
    my @row = $sql_get_mac->fetchrow_array();
    $sql_get_mac->finish();
    return $row[0];
}

#----------------------------------------------------------------------------\\
#
# Non-Exported Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# prepare_sql: Initialises all global SQL queries for use with the current
# database handle.
#
# Parameters: None.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub prepare_sql {

      #TODO: do select/insert by column name, not by location!

      $sql_log = $dbh->prepare("insert into interface_log ".
               "values (?,?,?,?,?,?,?, ".
               "?,?,?,?,?)");

      $sql_check_existing_mac = $dbh->prepare("select mac from interface ".
               "where mac = ?");

      $sql_update_mac = $dbh->prepare(
               "update interface set hostname = ?, ".
               "assetno = ?, owner = ?, domain = ?, status = ?, ".
               "statusdate = ?,platform = ?,comments=?,ip=?, ".
               "cidr_subnet=?,cidr_mask=?,group_id=? where mac = ?");

      $sql_insert_mac = $dbh->prepare("insert into interface ".
               " values (?, ?, ?, ?, ?, ?, ?,?,?,?,?,?,?,null)");

      $sql_expire_mac = $dbh->prepare("update interface set statusdate = ".
               "?, status='R' ".
               "where mac = ?");

      $sql_lock_mac = $dbh->prepare("update interface set status = 'U', ".
               "statusdate = ? where mac = ?");

      $sql_registered = $dbh->prepare("select mac,statusdate from interface ".
                                      "where mac = ? ");

      $sql_active = $dbh->prepare("select mac,statusdate from interface where ".
                                  "mac = ? and status = 'R'");

      $sql_locked = $dbh->prepare("select mac,statusdate from interface where ".
                                  "mac = ? and status = 'U'"); 

      $sql_unregip = $dbh->prepare("select unregip from interface_log ".
               "where mac = ? order by entrydate desc");

      $sql_getall = $dbh->prepare("select * from interface where mac = ?");

      $sql_unique_host = $dbh->prepare("select mac,statusdate from interface ".
             "where hostname = ? and domain = ?");

      $sql_static_all = $dbh->prepare("select * from interface where ".
               "status='S' and cidr_subnet = ? and cidr_mask = ?");

      $sql_del_static = $dbh->prepare("delete from interface where status='S' ".
                                      "and mac = ?");
     
      $sql_all_subnets = $dbh->prepare("select * from subnets");

      $sql_subnet_desc = $dbh->prepare("select * from subnets where ".
                                       "cidr_subnet = ? and cidr_mask = ?");

      $sql_get_campuses = $dbh->prepare("select distinct campus from campuses");

      $sql_group_list = $dbh->prepare("select * from interface where ".
                                      "group_id = ?");
      $sql_get_mac = $dbh->prepare("select mac from interface where ".
                                   "hostname = ? and domain = ?");
}
#----------------------------------------------------------------------------/
1;

