#----------------------------------------------------------------------------\
# NetReg 2.0 (pre-release devel version)
#       DHCPD Registration Frontend
#
# Version 1.3:
# Copyright (C) 1999-2002 Peter Valian, Southwestern University
# Version 2.0:
# Portions Copyright (C) 1999-2002 Peter Valian, Southwestern University.
# Portions Copyright (C) 2002 David Bussenschutt, Griffith University, Australia.
# Portions Copyright (C) 2002 Robert Lowe,  Lawrence University, Wisconsin.
# Portions Copyright (C) 2002-2003 Michael McMaster, Griffith University, Australia.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Version 1.3: http://www.netreg.org/
# Version 2.0: http://www.sourceforge.net/projects/netreg/
#
# Version 1.3 code contributions by:
#   Peter Valian <valianp@southwestern.edu>,  Southwestern University.
#     Author of Version 1.3rc2
#   Bobby Clark <bobby.clark@eku.edu>, Eastern Kentucky University
#     IMAP Authentication
#   Michael Grinnell <grinnell@american.edu>, American University
#     LDAP Authentication
#
# Version 2.0 code contributions by:
#   Robert Lowe <robert.h.lowe@lawrence.edu>, Lawrence University, Wisconsin.
#     CIDR (Subnet Masks etc) Handling/Notation
#     Alpha Testing, and bug chasing in David's Code.
#   David Bussenschutt <D.Bussenschutt@mailbox.gu.edu.au>, Griffith University, Queensland, Australia.
#     Much re-arrangement, and fixes, including but not limited to:
#     Perl 5 Syntax, and modules
#     File Locking
#     OMAPI Support
#     Remote dhcpdmanager daemon, and client support (functionally similar to omapi, but more fully implemented)
#     Improved Error Handling
#     Use of -w for improved code
#     LDAP group handling, for Permissions.
#     install_script
#   Michael McMaster <m4k3r@users.sourceforge.net>, Griffith University,  Queensland, Australia.
#     DBI/Oracle Support
#     HTML extraction/templating
#     Permissions handling (access controls)
#     Cookie based authentication
#     Interface for editing dhcp options/groups and hosts.
#     Synchronisation between hosts
#     Scalability improvements
#     Various code clean-ups
#     Consistency checker for host information
#----------------------------------------------------------------------------/
package Group;
use strict;
use warnings;

use Exporter();
our @ISA = qw(Exporter);

our @EXPORT = qw( );
our @EXPORT_OK =  qw(write_hosts);
use UNIVERSAL qw(isa);

#----------------------------------------------------------------------------\\
#
# Imported Modules
#
#----------------------------------------------------------------------------//
use Html       qw(print_html_file admin_error);
use Database   qw(get_group_list db_get_all);
use Subnets    qw(ip2dec);
use Admin      qw(get_color);
use Variables  qw($LDAP_URL $DDNS_USE_HOST_DECL $DEBUG);
use NetregMisc qw(make_time append_error_log);

#----------------------------------------------------------------------------\\
#
# Object Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# new: Creates a new "Group" object. (Stores info about dhcp groups).
#
# Parameters: The name of this group. eg. group {\n#name
#
# Returns: A group object (a blessed hash reference).
#----------------------------------------------------------------------------//
sub new {
    my $class = shift;
    my $name = shift;
    my $self  = {}; 

    $self->{'options'} = {};
    $self->{'hosts'} = {};
    $self->{'groups'} = {};
    $self->{'text'} = '';
    $self->{'contents'} = 0;

    $self->{'name'} = $name; 

    bless($self, $class);    
    return $self;
}   

#----------------------------------------------------------------------------\\
# display: Recursivly display the contents of this group as HTML.
#
# Parameters: The cumulatative group id string.
#             A saved md5 hash of the dhcp conf file
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub display {
    my $self = shift;
    my $url = shift;
    my $saved_md5 = shift;

    print_html_file("static_banner.html", "\\^TYPE", "SECTIONS");

    my $c_url = $url ? "$url:" : '';

    # Display Groups

    # Assume that group members are either ALL subnets or ALL not
    # subnets. If this is wrong, then the sort order will be out a bit,
    # which isn't really a big problem.
    my @all_groups = keys %{$self->{'groups'}};
    my @sorted_groups;
    if ($all_groups[0] && $all_groups[0] =~ /^[\d\.]+\/\d+$/) {
	if ($all_groups[$#all_groups] && $all_groups[$#all_groups] =~ /^[\d\.]+\/\d+$/) {
          @sorted_groups = sort ({ip2dec($a) <=> ip2dec($b)} @all_groups);
	} else {
	  @sorted_groups = sort @all_groups;
	}
    } else {
        @sorted_groups = sort @all_groups;
    }
    foreach (@sorted_groups) {
        print_html_file("static_group_row.html", "\\^CUMULATATIVE_URL", $c_url,
                            "\\^COMMENT", $_,
                            "\\^SAVED_MD5", $saved_md5, "\\^URL", $url);
    }

    # Display options

    my $test = 0;
    foreach (sort keys %{$self->{'options'}}) {
        print_html_file("static_banner.html", "\\^TYPE", "OPTIONS") unless $test;
        $self->{'options'}{$_}->display($url, $saved_md5);
        $test++;
    }

    # Display hosts

    my $colour = get_color();
    my @hosts = get_group_list($url);
    my %ip_host;
    $test = 0;
    foreach (@hosts) {
        $test++;
        my @row = db_get_all($_);

        #force dynamic hosts to appear after statics
        $ip_host{$row[5] eq 'S' ? ip2dec($row[9]) : "9$test"} = \@row;
    }
    if ($test) {
        print_html_file("static_banner.html", "\\^TYPE", "HOSTS");
        print_html_file("static_viewhost_start.html");
    }

    # Force a string comparison, not a numerical one.
    foreach (sort ({$a cmp $b} (keys %ip_host))) {
        my @row = @{$ip_host{$_}};
        my $userlink = $row[3]; # Default null link without LDAP.
        if ($LDAP_URL) {
            my $LDAP_U = $LDAP_URL;
            $LDAP_U =~ s/<USERNAME>/$row[3]/;
            $userlink = "<A HREF=$LDAP_U target=blank>$row[3]</A>";
        }
        my $exp = $row[5] eq 'S' ? '' : 'Expire';
        my $del = $row[5] eq 'S' ? 'Delete' : '';

        $row[9] = 'dynamic' if $row[5] ne 'S';
        print_html_file("static_host_row.html", "\\^COLOR", $colour,
                        "\\^HOST", "$row[1].$row[4]",
                        "\\^MAC", $row[0], "\\^IP", $row[9],
                        "\\^PLATFORM", $row[7],
                        "\\^OWNER", $userlink, "\\^ASSET", $row[2],
                        "\\^COMMENT", $row[8],
                        "\\^SAVED_MD5", $saved_md5, "\\^URL", $url,
                        "\\^EXPIRE", $exp, "\\^DELETE", $del);
        $colour = get_color($colour);
    }
    print "</table>\n" if $test;
}

#----------------------------------------------------------------------------\\
# write: write the contents of this Group to a given file handle in dhcpd.conf
# format.
#
# Parameters: A filehandle to write to.
#             An indentation level
#             The cumulative URL of this group. (eg. g1:g2:g3...)
#             True if hosts from the database should be inserted.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub write {
    my $self = shift;
    my $fh = shift;
    my $level = shift || 0;
    my $c_url = shift;
    my $hosts = shift;

    $self->{'name'} = '' unless $self->{'name'};

    $c_url .= ':' if $c_url;
    $c_url .= $self->{'name'};

    pad ($fh, $level - 1);
    print $fh "group {\n" if $level;
    pad ($fh, $level - 1);
    print $fh "#".$self->{'name'}."\n";

    my @lines = split('\n', $self->{'text'});
    foreach (@lines) {
        pad ($fh, $level);
        print $fh "$_\n";
    }

    foreach (sort keys %{$self->{'options'}}) {
        pad ($fh, $level);
        $self->{'options'}{$_}->write($fh);
    }

    write_hosts($fh, $c_url, $level) if $hosts;

    foreach (sort keys %{$self->{'groups'}}) {
        $self->{'groups'}{$_}->write($fh, $level+1,$c_url,$hosts);
    }
    print $fh "\n";
    pad ($fh, $level - 1);
    print $fh "}\n" if $level;
}

#----------------------------------------------------------------------------\\
# write: writes dhcpd.conf host entries for all hosts in a given group.
#
# Parameters: A filehandle to write to.
#             The cumulative URL of this group. (eg. g1:g2:g3...)
#             An indentation level
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub write_hosts {
    my ($fh,$url,$level) = @_;


    # get hosts from the db.
    my @macs = get_group_list($url);
    my $time = make_time();
    $time =~ s/^(\d+).*/$1/;

    append_error_log("write_hosts: url:$url level:$level\n") if $DEBUG>=3;

    foreach (@macs) {
        my @row = db_get_all($_);
        next unless $row[6] > $time;
        next if $row[5] eq 'U';
        
        if ($row[5] eq 'S') {
            pad ($fh, $level);
            print $fh "host $row[1].$row[4] { hardware ethernet $_ ; fixed-address ".
                      "$row[9] ; }\n";
            append_error_log("host $row[1].$row[4] { hardware ethernet $_ ; fixed-address ".
                      "$row[9] ; }\n") if $DEBUG>=3;
        } else {
            pad ($fh,$level);
            print $fh "host $row[1].$row[4] { hardware ethernet $_ ;}\n";
            append_error_log("host $row[1].$row[4] { hardware ethernet $_ ;}\n") if $DEBUG>=3;
            if ($DDNS_USE_HOST_DECL) {
                pad ($fh,$level);
                print $fh "subclass \"match-hardware\" 1:$_ { ".
                    "ddns-hostname \"$row[1]\"; option domain-name \"$row[4]\"; }\n";
                append_error_log("subclass \"match-hardware\" 1:$_ { ".
                    "ddns-hostname \"$row[1]\"; option domain-name \"$row[4]\"; }\n") if $DEBUG>=3;
            }
        }
    }
}

#----------------------------------------------------------------------------\\
# name: returns the name of this class.
#
# Parameters: None.
#
# Returns: The name of this class.
#----------------------------------------------------------------------------//
sub name {
    my $self = shift;
    $self->{'name'} = '' unless $self->{'name'};
    return $self->{'name'};
}

#----------------------------------------------------------------------------\\
# is_empty: returns true if this group does not contain any options, hosts,
# or sub-groups.
#
# Parameters: The cumulative URL of this group. (eg. g1:g2:g3...)
#
# Returns: The name of this class.
#----------------------------------------------------------------------------//
sub is_empty {
    my ($self,$url) = @_;

    my @macs = get_group_list($url);
    return $self->{'contents'} == 0 && $#macs == -1;
}

#----------------------------------------------------------------------------\\
# add_option: Adds a given option to this group if the option does not already
# exist.
#
# Parameters: An Option object to add.
#
# Returns:  Nothing useful.
#----------------------------------------------------------------------------//
sub add_option {
    my ($self,$option) = @_;
    $self->{'contents'}++ unless $self->{'options'}{$option->name};
    $self->{'options'}{$option->name} =  $option unless 
                      $self->{'options'}{$option->name};
}

#----------------------------------------------------------------------------\\
# update_option: Updates a given option name with a new value.
#   
# Parameters: The option name to update
#             A new value.
#
# Returns:  Nothing useful.
#----------------------------------------------------------------------------//
sub update_option {
    my ($self,$option_name,$option_val) = @_;
    $self->{'options'}{$option_name}->update($option_val) 
                      if $self->{'options'}{$option_name};
}

#----------------------------------------------------------------------------\\
# delete_option: Removes an option from this group.
#  
# Parameters: The option name to delete
#
# Returns:  Nothing useful.
#----------------------------------------------------------------------------//
sub delete_option {
    my ($self,$option_name) = @_;
    $self->{'contents'}-- if $self->{'options'}{$option_name};
    delete $self->{'options'}{$option_name};
}

#----------------------------------------------------------------------------\\
# add_group: Adds a subgroup to this group.
# 
# Parameters: A Group object to add.
#
# Returns:  Nothing useful.
#----------------------------------------------------------------------------//
sub add_group {
    my ($self,$group) = @_;
    $self->{'contents'}++ unless $self->{'groups'}{$group->name};
    $self->{'groups'}{$group->name} =  $group unless 
                     $self->{'groups'}{$group->name};
}

#----------------------------------------------------------------------------\\
# get_group: Returns the sub-Group object corresponding to the given group name.
#
# Parameters: A group name to find.
#
# Returns: A group object.
#----------------------------------------------------------------------------//
sub get_group {
    my ($self,$group_name) = @_;
    return $self->{'groups'}{$group_name};
}

#----------------------------------------------------------------------------\\
# get_groups: Returns an array of sub-group objects.
#   
# Parameters: None.
#   
# Returns: An array of group objects. 
#----------------------------------------------------------------------------//
sub get_groups {
    my $self = shift;
    return values %{$self->{'groups'}};
}

#----------------------------------------------------------------------------\\
# get_text: Returns any free-text strings in this group.
#  
# Parameters: None.
#  
# Returns: Some text in the dhcpd.conf format.
#----------------------------------------------------------------------------//
sub get_text {
    my $self = shift;
    return $self->{'text'};
}


#----------------------------------------------------------------------------\\
# delete_group: Deletes a subgroup from this group.
#
# Parameters: A group name to delete.
#             The cumulative URL of this group. (eg. g1:g2:g3...)
#
# Returns: Undef on failure.
#----------------------------------------------------------------------------//
sub delete_group {
    my ($self,$group_name,$url) = @_;
    return 1 unless $self->{'groups'}{$group_name};

    if ($self->{'groups'}{$group_name}->is_empty($url)) {
        $self->{'contents'}--;
        delete $self->{'groups'}{$group_name};
    } else {
        admin_error('Group is not empty!', "Consistency");
        return undef;
    }
}

#----------------------------------------------------------------------------\\
# update_text: Sets the free-form text (dhcpd.conf format) of this group.
#
# Parameters: Some text.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub update_text {
    my $self = shift;
    my $text = shift || '';

    # Some browsers send the string with MAC style carriage returns...
    # Time to fix this.
    $text =~ tr/\r/\n/;
    $text =~ s/(\s)\s*/$1/mg;

    $self->{'text'} = $text;
}

#----------------------------------------------------------------------------\\
#
# Non-Object Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# pad: prints a given number of spaces to a filehandle.
#  
# Parameters: A filehandle to print to.
#             The level of padding (4 times this number will be printed)
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub pad {
    my $fh = shift;
    my $level = 4 * shift;
    print $fh sprintf("%$level".'s','') if $level;
}

1;
