#----------------------------------------------------------------------------\
# NetReg 2.0 (pre-release devel version)
#       DHCPD Registration Frontend
#
# Version 1.3: 
# Copyright (C) 1999-2002 Peter Valian, Southwestern University
# Version 2.0:
# Portions Copyright (C) 1999-2002 Peter Valian, Southwestern University.
# Portions Copyright (C) 2002 David Bussenschutt, Griffith University, Australia.
# Portions Copyright (C) 2002 Robert Lowe,  Lawrence University, Wisconsin.
# Portions Copyright (C) 2002-2003 Michael McMaster, Griffith University, Australia.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Version 1.3: http://www.netreg.org/
# Version 2.0: http://www.sourceforge.net/projects/netreg/
#
# Version 1.3 code contributions by:
#   Peter Valian <valianp@southwestern.edu>,  Southwestern University.
#     Author of Version 1.3rc2 
#   Bobby Clark <bobby.clark@eku.edu>, Eastern Kentucky University 
#     IMAP Authentication
#   Michael Grinnell <grinnell@american.edu>, American University
#     LDAP Authentication
#
# Version 2.0 code contributions by:
#   Robert Lowe <robert.h.lowe@lawrence.edu>, Lawrence University, Wisconsin.
#     CIDR (Subnet Masks etc) Handling/Notation
#     Alpha Testing, and bug chasing in David's Code.
#   David Bussenschutt <D.Bussenschutt@mailbox.gu.edu.au>, Griffith University, Queensland, Australia.
#     Much re-arrangement, and fixes, including but not limited to:
#     Perl 5 Syntax, and modules
#     File Locking
#     OMAPI Support
#     Remote dhcpdmanager daemon, and client support (functionally similar to omapi, but more fully implemented)
#     Improved Error Handling
#     Use of -w for improved code
#     LDAP group handling, for Permissions.
#     install_script
#   Michael McMaster <m4k3r@users.sourceforge.net>, Griffith University,  Queensland, Australia.
#     DBI/Oracle Support
#     HTML extraction/templating
#     Permissions handling (access controls)
#     Cookie based authentication
#     Interface for editing dhcp options/groups and hosts.
#     Synchronisation between hosts
#     Scalability improvements
#     Various code clean-ups
#     Consistency checker for host information 
#----------------------------------------------------------------------------/
package NetregMisc;
use strict;
use warnings;
#----------------------------------------------------------------------------\\
#
# Exports
#
#----------------------------------------------------------------------------//
use Exporter();
our @ISA = qw(Exporter);

our @EXPORT = qw( );
our @EXPORT_OK = qw ( make_time append_error_log open_with_lock append_import_log ); 

#----------------------------------------------------------------------------\\
#
# Imported Modules
#
#----------------------------------------------------------------------------//
# If using mod_perl, the $FindBin::RealBin directory will be set ONCE only
# see startup.pl
use FindBin;
use lib "$FindBin::RealBin"; # locate library path for below modules.

use Fcntl qw(:DEFAULT :flock); # for file locking

# Netreg modules
use Variables qw ( $ERROR_LOG $IMPORT_LOG );

#----------------------------------------------------------------------------\\
#
# Constants
#
#----------------------------------------------------------------------------//
# Optional mod_perl support.
# Sets a constant, and ensures the correct exit() function is called.
use constant IS_MODPERL => $ENV{MOD_PERL};
use subs qw(exit);
# Select the correct exit function
*exit = IS_MODPERL ? \&Apache::exit : sub { CORE::exit };

#----------------------------------------------------------------------------\\
#
# Exported Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# append_error_log: Sends an error message to a log file.
#
# Parameters: An array of error messages
#
# Returns: Nothing useful
#----------------------------------------------------------------------------//
sub append_error_log {
    my @data = @_;

    my $error_log = open_with_lock(">> $ERROR_LOG",1);
    if (!$error_log) {
        print STDERR "Cannot open log file $ERROR_LOG $!\n";
        return undef;
    }

    # print a timestamp to the file
    my $timestamp = make_time();
    print $error_log "$timestamp :", @data, "\n";

    close $error_log;
    return 1;
}


#----------------------------------------------------------------------------\\
# append_import_log: Sends an import message to a log file.
#
# Parameters: An array of import data batches
#
# Returns: Nothing useful
#----------------------------------------------------------------------------//
sub append_import_log {
    my @data = @_;

    my $import_log = open_with_lock(">> $IMPORT_LOG",1);
    if (!$import_log) {
        print STDERR "Cannot open log file $IMPORT_LOG $!\n";
        return undef;
    }

    # print a timestamp to the file
    my $timestamp = make_time();
    print $import_log "$timestamp :", @data, "\n";

    close $import_log;
    return 1;
}


#----------------------------------------------------------------------------\\
# open_with_lock: Opens the given file with an exclusive lock. This may block
# while attempting to lock.
#
# Parameters: The filename to open (complete with "<" or "> modifiers)
#             True if errors shouldn't be logged (ONLY necessary when called
#                         from the logging functions themselves)
#
# Returns: A filehandle, or undef on error
#----------------------------------------------------------------------------//
sub open_with_lock {
    my $filename = shift;
    my $nolog = shift;

    print STDERR "open_with_lock: attempting to lock $filename\n" unless $nolog;

    if (!open(LOCK_FILE, $filename)) {
      append_error_log ( "Can not open $filename : $! \n") unless $nolog;
      return undef;
    }
    unless (flock(LOCK_FILE, LOCK_EX | LOCK_NB)) {
      local $| = 1;  # turn on autoflush
      if (!flock(LOCK_FILE, LOCK_EX) ) {
        append_error_log ("Can't lock $filename : $!\n") unless $nolog;
        return undef;
      }
    }

    print STDERR "open_with_lock: successfully locked $filename\n" unless $nolog;

    return \*LOCK_FILE;
}

#----------------------------------------------------------------------------\\
# make_time: Returns a timestamp (as a string) for the current time.
#
# Parameters: none.
#
# Returns: A 'YYYYMMDD HH:MM:SS' timestamp
#----------------------------------------------------------------------------//
sub make_time {

    my ($s,$m,$h,$dy,$mo,$yr,$wd,undef,undef) = localtime;
    return sprintf("%02d%02d%02d %02d:%02d:%02d", $yr+1900,$mo+1,$dy,$h,$m,$s);
}

#----------------------------------------------------------------------------/
1;

