#----------------------------------------------------------------------------\
# NetReg 2.0 (pre-release devel version)
#       DHCPD Registration Frontend
#
# Version 1.3:
# Copyright (C) 1999-2002 Peter Valian, Southwestern University
# Version 2.0:
# Portions Copyright (C) 1999-2002 Peter Valian, Southwestern University.
# Portions Copyright (C) 2002 David Bussenschutt, Griffith University, Australia.
# Portions Copyright (C) 2002 Robert Lowe,  Lawrence University, Wisconsin.
# Portions Copyright (C) 2002 Michael McMaster, Griffith University, Australia.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Version 1.3: http://www.netreg.org/
# Version 2.0: http://www.sourceforge.net/projects/netreg/
#
# Version 1.3 code contributions by:
#   Peter Valian <valianp@southwestern.edu>,  Southwestern University.
#     Author of Version 1.3rc2
#   Bobby Clark <bobby.clark@eku.edu>, Eastern Kentucky University
#     IMAP Authentication
#   Michael Grinnell <grinnell@american.edu>, American University
#     LDAP Authentication
#
# Version 2.0 code contributions by:
#   Robert Lowe <robert.h.lowe@lawrence.edu>, Lawrence University, Wisconsin.
#     CIDR (Subnet Masks etc) Handling/Notation
#     Alpha Testing, and bug chasing in David's Code.
#   David Bussenschutt <D.Bussenschutt@mailbox.gu.edu.au>, Griffith University, Queensland, Australia.
#     Much re-arrangement, and fixes, including but not limited to:
#     Perl 5 Syntax, and modules
#     File Locking
#     OMAPI Support
#     Remote dhcpdmanager daemon, and client support (functionally similar to omapi, but more fully implemented)
#     Improved Error Handling
#     Use of -w for improved code
#     LDAP group handling, for Permissions.
#     install_script
#   Michael McMaster <m4k3r@users.sourceforge.net>, Griffith University,  Queensland, Australia.
#     optional DBI/Oracle Support (still works without the DB, but is much quicker with it when using remote servers)
#     HTML extraction/templating
#     Permissions handling (two levels of access)
#     Login handling (cookie based authentication)
#     Interface for editing details of static hosts
#     Synchronisation between hosts
#     Improvement of the scalability of the overall system
#     Cleaning up most existing code
#----------------------------------------------------------------------------\
package StaticHosts;
use strict;
use warnings;

#----------------------------------------------------------------------------\\
#
# Exports
#
#----------------------------------------------------------------------------//
use Exporter();
our @ISA = qw(Exporter);

our @EXPORT = qw( );
our @EXPORT_OK =  qw( print_static_page update_static add_static_parameter 
                      delete_static get_cnames parse_conf write_conf get_groups
                      get_js_groups get_top_groups get_groups_hash get_group_indexes );


#----------------------------------------------------------------------------\\
#
# Imported Modules
#
#----------------------------------------------------------------------------//
# If using mod_perl, the $FindBin::RealBin directory will be set ONCE only
# see startup.pl
use FindBin;
use lib "$FindBin::RealBin"; # locate library path for below modules.
use Socket;
use Data::Dumper;


# Netreg modules.
use Variables qw( $STATIC_FILE $DHCPDMASTERLOCK $ADMINPATH $Key $KeyName $DHCPDBINARY 
                  $LDAP_URL $NSUPDATE $DIG @ZONES $DHCPOPTIONSFILE @ZONEFILES $FGREP $DEBUG );
use Authenticate qw (user priv);
use Html qw(get_input register_error admin_error print_html_file);
use Database qw (db_get_all db_unique_host db_new_host db_next_serial 
                 db_del_static  db_active db_get_mac db_mac_locked);
use Subnets qw(read_domains ip2dec what_is_my_registered_subnet add_new_subnet_to_system );
use NetregMisc qw (make_time append_error_log );
use Group;
use Option;
use SubnetDec;
use Pool;
use Class;
use SharedNet;

#----------------------------------------------------------------------------\\
#
# Constants
#
#----------------------------------------------------------------------------//
# Optional mod_perl support.
# Sets a constant, and ensures the correct exit() function is called.
use constant IS_MODPERL => $ENV{MOD_PERL};
use subs qw(exit);
# Select the correct exit function
*exit = IS_MODPERL ? \&Apache::exit : sub { CORE::exit };

#----------------------------------------------------------------------------\\
#
# Global Variables
#
# These are shared within this perl module only.
# These variables are specifically shared between executions if using mod_perl
#----------------------------------------------------------------------------//

# The saved values of the dhcpd master file.
my $saved_group; 

# The mtime of the dhcpd master file when it was saved to the %saved_hash
# variable. This used to be a md5 hash, but was changed for performance reasons.
my $saved_md5 = '';

# An array of dhcpd settings that should act as options.
my %dhcp_options;

#----------------------------------------------------------------------------\\
#
# Exported Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# update_static: Updates existing details in the dhcpd master file. Does not
# include host information. See add_static_parameter
#
# Parameters: A filehandle of a temporary file.
#
# Returns: '1' on success, undef on error.
#----------------------------------------------------------------------------//
sub update_static {
    my $filehandle = shift;
    my %form = get_input();

    my $static_hosts = parse_conf($form{'hash'});
    return undef unless $static_hosts;

    my $display_hash = get_section($form{'section'}, $static_hosts);
    if (!$display_hash) {
        admin_error("DHCPD information undefined. Please try again. ".
                    "Was the web server restarted?","Debug");
        return "1";
    } 

    # Get privileges.
    if (!priv("group_".$form{'section'})) {
        admin_error("Insufficient Privileges. No access to this section.", "Privilege");
        print_html_file("admin_footer.html");
    }

    # Deal with the update, depending on what is being updated. 
    if ($form{'parameter'} eq 'option') {
        if (!priv('option_'.$form{'option'})) {
            admin_error("Insufficient Privileges", "Privilege");
            print_html_file("admin_footer.html");
        }
        $display_hash->update_option($form{'option'},$form{'optval'});

    } elsif ($form{'parameter'} eq 'text') {
        if (!priv('edit_dhcp_text')) {
            admin_error("Insufficient Privileges", "Privilege");
            print_html_file("admin_footer.html");
        }
        $display_hash->update_text($form{'optval'});
    } else {
        admin_error("Unsupported operation", "Debug");
        return "0";
    }
    write_conf($static_hosts, $filehandle);
}

#----------------------------------------------------------------------------\\
# delete_static: Deletes an entry from the dhcpd master file.
#
# Parameters: A filehandle of a temporary file.
#
# Returns: '1' on success, undef on error.
#----------------------------------------------------------------------------//
sub delete_static {
    my $filehandle = shift;
    my %form = get_input();
    my $static_hosts = parse_conf($form{'hash'});
    return undef unless $static_hosts;

    my $display_hash = get_section($form{'section'}, $static_hosts);
    if (!$display_hash) {
        admin_error("DHCPD information undefined. Please try again. ".
                    "Was the web server restarted?","Debug");
        return "1";
    }

    # Get privileges.
    if (!priv("group_".$form{'section'})) {
        admin_error("Insufficient Privileges", "Privilege");
        print_html_file("admin_footer.html");
    }

    # Delete the entry, depending on it's type.
    if ($form{'parameter'} eq 'option') {
        if (!priv('option_'.$form{'option'})) {
            admin_error("Insufficient Privileges", "Privilege");
            print_html_file("admin_footer.html");
        }
        $display_hash->delete_option($form{'option'});

    } elsif ($form{'parameter'} eq 'host') {
        if (!priv('edit_static_hosts')) {
            admin_error("Insufficient Privileges", "Privilege");
            print_html_file("admin_footer.html");
        }
        $form{'host'} =~ /^([a-z0-9\-]+)\.(.+)$/i;
        my $mac = db_get_mac($1,$2); 

        my @row = db_get_all($mac);
        if ($row[5] ne 'S') {
            admin_error("Please use the expire function to delete dynamic hosts");
            print_html_file("admin_footer.html");
        }
        
        my @cnames = get_cnames($form{'host'});
        my $all = $form{'all'};
        my $update_dns = 1; # Change dns entries by default.
        my $fqdn = "$row[1].$row[4]";

        my $packed_ip = inet_aton($row[9]);
        if (gethostbyaddr($packed_ip, AF_INET) ne $fqdn) {
            admin_error("Host/IP are not consistent in the DNS system. DNS not updated,".
                "for $fqdn/$row[9]. Please do so manually.", "Static DNS");
            $update_dns = 0;
        }

        # Check to see if the host details are consistent
        my $dns_ip = gethostbyname($fqdn);
        if ($dns_ip && ($row[9] ne inet_ntoa($dns_ip))) {
                admin_error("Host/IP are not consistent in the DNS system. DNS not updated,".
                            "for $fqdn/$row[9]. Please do so manually.", "Static DNS");
                $update_dns = 0;
        }

        if ($update_dns) {
            my $reverse = reverse_ip($row[9]);
            open (NSUPDATE, "| $NSUPDATE -y $KeyName:$Key");
            print NSUPDATE "update delete $fqdn";
            print NSUPDATE " A" if !$all;
            print NSUPDATE "\n\n";
            print NSUPDATE "update delete $reverse\n\n";

            if ($all) {
                foreach (@cnames) {
                    print NSUPDATE "update delete $_\n\n";
                }
            }
            close NSUPDATE;
        }
        db_del_static($row[0]);

    } elsif ($form{'parameter'} eq 'group') {
        if (!priv('edit_dhcp_group')) {
            admin_error("Insufficient Privileges", "Privilege");
            print_html_file("admin_footer.html");
        }

        if (!$display_hash->delete_group($form{'option'},$form{'section'})) {
            return 1;
        }

    } else {
        admin_error("Unsupported operation", "Debug");
        return "0";
    }
    write_conf($static_hosts, $filehandle);
}

#----------------------------------------------------------------------------\\
# get_cnames: Returns an array of cnames for the given host. A zone file is
# grep'd for the information.
#
# Parameters: A fully qualified domain name.
#
# Returns: An array of cnames for the host.
#----------------------------------------------------------------------------//
sub get_cnames {
    my $fqdn = shift;

    $fqdn =~ /^([a-z0-9\-]+)/i;
    my $host = $1;
    my @cnames = ();

    foreach (@ZONEFILES) {
        # This next line requires external utilities
        # but is FAST for large zone files.
        my @lines = `$FGREP -e '$host' -e 'ORIGIN' $_`;
        
        my $domain = '';
        foreach my $line (@lines) {
            if ($line =~ /^\$ORIGIN ([a-z0-9\-\.]+)\./i) {
                $domain = $1;
                next;
            }
     
            if ($line =~ /^([a-z0-9\-\.]+).*CNAME\s+([a-z0-9\-\.]+)/i) {
                my $match = "$2.$domain";
                my $cname = $1;
                if ($match =~ /^$fqdn\.*$/) {
                    $cname .= ".$domain" if $domain;
                    push @cnames, $cname;
                }
            } 
        }
    }
    return @cnames;
}

#----------------------------------------------------------------------------\\
# add_static_parameter: Adds an entry to the dhcpd master file, including
# options, groups, static and roaming hosts, and subnets (ignore the function name).
# Also handles updating of host information (ie. NOT update_static)
#
# Parameters: A filehandle of a temporary file.
#
# Returns: '1' on success, undef on error.
#----------------------------------------------------------------------------//
sub add_static_parameter {
    my $filehandle = shift;
    my %form = get_input();
    my $static_hosts = parse_conf($form{'hash'});
    return undef unless $static_hosts;

    if (!($form{'id'} || $form{'mac'})) {
        admin_error("Missing information!", "Input");
        return;
    }

    # "section" isn't defined when adding a NEW section...
    $form{'section'} = '' unless defined $form{'section'};

    # Get privileges.
    if (!priv("group_".$form{'section'})) {
        admin_error("Insufficient Privileges to section.", "Privilege");
        print_html_file("admin_footer.html");
    }

    my $display_hash = get_section($form{'section'}, $static_hosts);
    if (!$display_hash) {
        admin_error("DHCPD information undefined. Please try again. ".
                    "Was the web server restarted?","Debug");
        return "1";
    }
    
    # Add the info, depending on its type.
    if ($form{'parameter'} eq 'option') {
        if (!priv('option_'.$form{'id'})) {
            admin_error("Insufficient Privileges", "Privilege");
            print_html_file("admin_footer.html");
        }
        my $option;
        if ($dhcp_options{$form{'id'}}) {
            $option = Option->new($form{'id'}, $form{'optval'},1);
        } else {
            $option = Option->new($form{'id'}, $form{'optval'});
        }
        $display_hash->add_option($option);

    } elsif ($form{'parameter'} eq 'group') {
        if (!priv('edit_dhcp_group')) {
            admin_error("Insufficient Privileges", "Privilege");
            print_html_file("admin_footer.html");
        }
        my $group = Group->new($form{'id'});
        $display_hash->add_group($group);

    } elsif ($form{'parameter'} eq 'subnet') {
        if (!priv('edit_dhcp_group')) {
            admin_error("Insufficient Privileges", "Privilege");
            print_html_file("admin_footer.html");
        }
	if ( $form{'id'} && $form{'cidrmask'} && $form{'description'} 
	      && $form{'campus'} && $form{'domains'} ) {

	  #add the subnet here!...
	  add_new_subnet_to_system( $form{'section'},
		$form{'campus'},$form{'cidrmask'},$form{'id'},
		$form{'description'},$form{'domains'}
	  );
	  # HACK HERE!
	  # by doing a 'return' I am preventing the write_conf at the end of 
	  # this routine from being called.  The write_conf does a syntax check of the
	  # the file, which will fail because the above add_new_subnet_to_system call 
	  # does the writing of the config file itself in a short-circuit way.
	  return "1";
		
	} else {
	   #admin_error("Insufficient number of parameters to create subnet","Programming");
	   admin_error("Insufficient number of parameters to create subnet $form{'id'} $form{'cidrmask'} $form{'description'} $form{'campus'} $form{'domains'}","Programming");
            print_html_file("admin_footer.html");
	}

    } elsif ($form{'parameter'} eq 'host') {
        my $dynamic; 
        $dynamic = 1 if ($form{'dynamic'} eq 'true');
        my $priv_string = 'edit_static_hosts';
        $priv_string = 'edit_dynamic_hosts' if $dynamic;
        if (!priv($priv_string)) {
            admin_error("Insufficient Privileges. Failed $priv_string.", "Privilege");
            print_html_file("admin_footer.html");
        }
 
        if (db_mac_locked($form{'mac'})) {
            admin_error("Host is currently locked out of the system. Please expire the ".
                        "lock using the search hosts page first.");
            print_html_file("admin_footer.html");
        }

        my $expiry = $form{'expiry'};
        if (!$expiry) {
            $expiry = make_time();
            $expiry =~ /^(\d\d\d\d)(\d\d)(\d\d).*$/;
            my ($day,$month,$year) = ($3,$2,$1+5);
            $expiry = "$year$month$day";
        } else {
            $expiry =~ /(\d+).(\d+).(\d+)/;
            my $year = $3 > 2030 ? 2030 : $3;
            $expiry = "$year$2$1";
        }

        ($form{'id'},my $temp) = check_input();
        return 1 unless $temp;
        my $fqdn = $form{'id'}.".".$form{'domain'};
        $form{'mac'} = uc($temp);

        # Check to see if the user is trying to update from
        # existing DNS...
        my @host = db_get_all($form{'mac'});
        $priv_string = 'edit_static_hosts';
        $priv_string = 'edit_dynamic_hosts' if $host[5] && $host[5] ne 'S';
        if ($host[0] && !priv($priv_string)) {
            admin_error("Insufficient Privileges. Failed $priv_string on existing host.",
                        "Privilege");
            print_html_file("admin_footer.html");
        }
        my ($existing_host, $existing_domain, $existing_ip,$existing_fqdn);

        if ($host[0]) {
            $existing_ip = $host[9];
            $existing_host = $host[1];
            $existing_domain = $host[4];
            $existing_fqdn = "$host[1].$host[4]";
        }
       
        if (!$dynamic && $existing_ip && ($form{'ip'} eq $existing_ip) && $form{'update'}) {
            my $dns_ip = gethostbyname($fqdn);
            if ($dns_ip) {
                if ($existing_ip ne inet_ntoa($dns_ip)) {
                    admin_error("DNS entries not consistent with IP. Host has been ".
                                "updated to reflect this.", "No");
                    $form{'ip'} = inet_ntoa($dns_ip);
                }
            }
        }

        # Check that this isn't already allocated (the ip)
        if (!$dynamic) {
            my $name = gethostbyaddr(inet_aton($form{'ip'}), AF_INET);        
            if ($name && ($name ne $fqdn)) {
                admin_error (pop_up("IP in use by $name"), "Static DNS");
                return '1';
            }
        }

        # NOTE: NSUPDATE is opened TWICE, just in case
        # It fails to delete. This causes the thing to exit,
        # and hosts therefore don't get added.
        if (!$dynamic && $form{'update'}) {
            open (NSUPDATE, "| $NSUPDATE -y $KeyName:$Key");
            my $reverse = reverse($existing_ip);
            print NSUPDATE "update delete $fqdn A\n\n";
            print NSUPDATE "update delete $reverse\n\n";
            close NSUPDATE;
        }

        if (!$dynamic) {
            my $reverse = reverse_ip($form{'ip'});
            open (NSUPDATE, "| $NSUPDATE -y $KeyName:$Key");
            print NSUPDATE "update add $fqdn 3600 A $form{'ip'}\n\n";
            print NSUPDATE "update add $reverse 3600 PTR $fqdn\n\n";
            close NSUPDATE;
        }

        # Add MX records
        if ($form{'mx'} && priv('super')) {
            my @records = split(',', $form{'mx'});
            my $command = "update delete $fqdn MX\n\n";
            foreach (@records) {
                /^(\d+)\s+([a-z\-0-9\.]+)$/;
                $command .= "update add $fqdn 3600 MX $1 $2\n\n";
            }
            open (NSUPDATE, "| $NSUPDATE -y $KeyName:$Key");
            print NSUPDATE $command;
            close NSUPDATE;
        }

        # Add CNAME records.
        if ($form{'cname'} && priv('super')) {
            # Delete existing CNAMES
            my @cnames = get_cnames($fqdn);
            open (NSUPDATE, "| $NSUPDATE -y $KeyName:$Key");
            foreach (@cnames) {
               print NSUPDATE "update delete $_ CNAME\n\n";
            } 

            # Add new CNAME's
            my @records = split(',', $form{'cname'});
            my $command = '';
            foreach (@records) {
                $command .= "update add $_ 3600 CNAME $fqdn\n\n";
            }
            print NSUPDATE $command;
            close NSUPDATE;
        }

        $form{'group'} =~ s/:+$//;
        $form{'group'} = 'Roaming:Default' unless $form{'group'};

        $form{'asset'} = -1 if ($form{'nonasset'} =~ /true/);
        my $letter = 'S';
        $letter = 'R' if $dynamic;
        # update the database
        my ($subnet, $mask) = what_is_my_registered_subnet($form{'ip'});
        db_new_host($form{'mac'},$form{'id'},$form{'asset'},$form{'owner'},
                    $form{'domain'},$letter,$expiry, user(), $form{'platform'},
                    $form{'ip'}, $form{'comment'},$subnet,$mask,$form{'group'},
                    $form{'update'});
    } else {
        admin_error("Unsupported operation", "Debug");
        return "0";
    }
    write_conf($static_hosts, $filehandle);
    return "1";
}


#----------------------------------------------------------------------------\\
# print_static_page: Displays the contents of a section in the dhcpd conf
# master file.
#
# Parameters: A url/section in the form of section:groupid:groupid...
#
# Returns: Nothing
#----------------------------------------------------------------------------//
sub print_static_page {
    my $url = shift || '';
    # URLS in the form of "section:groupNO:groupNO"

    append_error_log("print_static_page: start") if $DEBUG>=1;

    my $static_hosts = parse_conf(); 

    # Print a navigation link thing.
    print "<h3><a href=$ADMINPATH/admin.cgi?action=view-static&section=>Home</a>";
    my $cumulatative_url = ''; # NOTE: this is saved for later.
    my @display_group = split(':',$url);
    foreach (@display_group) {
        $cumulatative_url .= $_;
        print "-&gt;<a href=$ADMINPATH/admin.cgi?action=view-static&section=".
              "$cumulatative_url>$_</a>";
        $cumulatative_url .= ':';
    }
    print "</h3>";

    append_error_log("print_static_page: mid") if $DEBUG>=1;

    # Get a hash of the content that we're interested in displaying.
    my $display_hash = get_section($url,$static_hosts);
    if (!$display_hash) {
        admin_error("Group does not exist","Config");
        print_html_file("admin_footer.html");
    }

    print_html_file("static_start.html", "\\^URL", $url,
                    "\\^SAVED_MD5", $saved_md5);

    # DISPLAY HERE
    $display_hash->display($url, $saved_md5);

    append_error_log("print_static_page: end") if $DEBUG>=1;

    print "<tr><td>&nbsp;</td></tr>\n";
    print_html_file("static_end.html", "\\^SAVED_MD5", $saved_md5,
                    "\\^URL", $url, "\\^STRING", $display_hash->get_text());
}

#----------------------------------------------------------------------------\\
# write_conf: Write the given hash to a dhcpd.conf.master file, but check
# the file for correctness first.
#
# Parameters: A hash (as returned by parse_conf) for writing.
#             A temporary filehandle to write to.
#
# Returns: Undef on error
#----------------------------------------------------------------------------//
sub write_conf {
    my $conf = shift;
    my $filehandle = shift ; # A temporary handle!

    $conf->write($filehandle);

    # Check the temp config file
    my $result = `$DHCPDBINARY -t -cf $DHCPDMASTERLOCK 2>&1`;

    $result = "Unable to Check Syntax with dhcpd command:\"$DHCPDBINARY -t -cf $DHCPDMASTERLOCK\"\n" unless $result;

    if (($result =~ /Configuration file errors encountered/ms) || ($result =~ /^exiting.$/ms)) {
    	$result =~ s/(errors encountered).*$/$1/ms;
        admin_error("Errors in dhcpd.conf format.<br />File not written.<br />".
                    "<div align=left><pre>$result</pre></div>","Dhcpd.conf");
        # wipe the saved hash
        $saved_md5 = 0;
    } elsif ($result =~ /Unable to Check Syntax with/ms) {
        admin_error("<br />Config file not written.<br />".
                    "<div align=left><pre>$result</pre></div>","Dhcpd.conf");
        # wipe the saved hash
        $saved_md5 = 0;
    } else {
        # The file works, so it's safe to write
        `cp $DHCPDMASTERLOCK $STATIC_FILE`;
    }
    return "1";
}

#----------------------------------------------------------------------------\\
# get_groups_hash: returns a hash of hashes of hashes representing the 
# group heirarchy in the dhcpd.conf.master file.
# 
# parameters: none
#
# returns: reference to an anonymous hash of hashes of hashes etc...
#----------------------------------------------------------------------------//
sub get_groups_hash {

my $HOH = {};
my $depth = 0;
my @output = ();
my @g = get_groups();

foreach my $grouplist (@g) {

  my $tmp = $HOH;
  my @groups = split(/:/,$grouplist);
  $depth = $#groups if $#groups > $depth;

  for ( my $i = 0 ; $i <= $#groups ; $i++ ) {
    $tmp->{$groups[$i]} = {} unless $tmp->{$groups[$i]};
    $tmp = $tmp->{$groups[$i]};
  }
}
# interested to see the structure? Use this:
#print Data::Dumper->Dump([$HOH]);

return $HOH;
}

#----------------------------------------------------------------------------\\
# gets indexes into the different groups for assistance in generating some of
# the javascript. 
# prerequisite... get_js_groups must have been called previously.
# parameters: none.
# returns: a hash where the key in each hash is the js array name, and 
#          the value in each hash is the heirarchical order in the 
#	   get_js_groups array where it appears.
#----------------------------------------------------------------------------//
my %get_group_index = ();
sub get_group_indexes {
 return %get_group_index;
}

#----------------------------------------------------------------------------\\
# get_js_groups: Generate a string of javascript arrays representing the group
# structure/heirarchy given by get_groups_hash, which calls get_groups, 
# which in turn calls parse_conf.
#
# Parameters: None.
#
# Returns: A string of javascript.
#----------------------------------------------------------------------------//

sub get_js_groups {
my $tmp_hash = get_groups_hash();

%get_group_index = (); # support mod_perl by clearing this here...

my $output = '';
$output .= "//----The following array is dynamically generated by get_js_groups----\n";
$output .= "array1 = new Array(\"\"";
$output = get_js_groups_recurse($output,$tmp_hash, 1 , '');
$output .= "\n);\n";
$output .= "//----The previous array was dynamically generated by get_js_groups---\n";

return $output;
}


#----------------------------------------------------------------------------\\
#utility function used by get_js_groups to actually make javascript, recursively
# as a side-effect, it also stores data into @get_group_indexes for the 
# subroutine 'get_group_indexes' to use.
#----------------------------------------------------------------------------\\
sub  get_js_groups_recurse {
  my $output = shift;
  my $t = shift;
  my $level = shift; # current depth in the tree
  my $path = shift;  # string representing the path we took to get to here in the tree.

  my $tabs = "\t" x $level;

  # if there are no keys, return 'null'
  my @keys =  keys %$t;
  if ($#keys < 0 ) {
    $output .= ",\n".$tabs."null";
    return $output;
  }

  # add the dummy entry before the real ones..
  $output .= ",\n".$tabs."new Array(\"\",\n";
  $output .= $tabs."\tnull\n";
  $output .= $tabs.")";

  # add the real ones...
  my  $j = 1; # start at 1 to skip the dummy entries.
  foreach (sort keys %$t) {
    # store this info for get_group_indexes..
    $j++;
    my $newpath = $path.','.$j;
    $path =~ s/^,*//;
    my @list = split /,/,$path;
    $get_group_index{$_}  = '['.join('][',@list)."]" if $#list >=0;
    $get_group_index{$_} .= '['.$j.']';
   
    # do the real stuff... 
    $output .= ",\n".$tabs."new Array(\"".$_."\"";
    $output = get_js_groups_recurse($output ,$$t{$_}, $level+1,$newpath);
    $output .= "\n".$tabs.")";
  }
  return $output;
}


#----------------------------------------------------------------------------\\
# get_top_groups: returns a count of how many sublevels there are in 
# the groups heirarchy...
#
# Parameters: None.
#
# Returns: a number.
#----------------------------------------------------------------------------//
sub get_top_groups {
    my @g = get_groups();
    my $ret = 0; 
    foreach my $g (@g) {
       my @t = split(/:/,$g);
       $ret=$#t+1 if $#t+1 >= $ret;
    }
    return $ret;
}

#----------------------------------------------------------------------------\\
# get_groups: Recursively returns an array of groups given by parse_conf.
# Multiple levels of groups are separated using :
#
# Parameters: Parameters are used internally by the function for recursive
# features. DO NOT pass it any parameters.
#
# Returns: An array of group names.
#----------------------------------------------------------------------------//
sub get_groups {
    my $current_group = shift || '';
    my $c_url = shift || '';

    $c_url .= ":" if $c_url;
    $current_group = parse_conf() unless $current_group;
    $c_url .= $current_group->name();

    my @groups = ();
    push @groups,($c_url) if $c_url;

    my @recursive = $current_group->get_groups();
    foreach (@recursive) {
        my @recursive_groups = get_groups($_, $c_url);
        push @groups, @recursive_groups;
    }
    
    return @groups;
}

#----------------------------------------------------------------------------\\
#
# Non-Exported Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# parse_conf: Create a hash from the dhcpd.conf.master file
#
# Parameters: An mtime (called a md5 hash for historical reasons)  of a previous
#             parse_conf call (optional)
#
# Returns: A hash reference, or undef if the given md5/mtime doesn't match the
#          current md5/mtime (if an md5/mtime hash was provided).
#----------------------------------------------------------------------------//
sub parse_conf {
    my $compare_hash = shift;
    my @contents;

    if (!%dhcp_options) {
        open (DHCP_OPTIONS, "<$DHCPOPTIONSFILE") || die 
                                        "Could not open $DHCPOPTIONSFILE. $!\n";
        while (my $line = <DHCP_OPTIONS>) {
            chomp($line);
            next if $line =~ /^#/;
            next if $line =~ /^\s*$/;
            $dhcp_options{$line}++;
        }
        close DHCP_OPTIONS;
    }

    # Get a hash of the file for consistency purposes
    open READCONF, "<$STATIC_FILE" || die "Could not open $STATIC_FILE. $!\n";
    @contents = <READCONF>;
    close READCONF;
    
    my $hash = (stat($STATIC_FILE))[9];
    if ($compare_hash) {
        if ($compare_hash eq $hash && $compare_hash eq $saved_md5) {
            return $saved_group;
        } elsif ($compare_hash ne $hash) {
            return undef;
        } else {
            # The given hash and the file match, BUT we need to
            # re-read the file 'cause we don't have the latest copy.
        }
    } elsif ($hash eq $saved_md5) {
        return $saved_group;
    }
    $saved_md5 = $hash;
       
    my ($group, $junk) = parse_section(\@contents);
    $saved_group = $group;

    return $group;
}

#----------------------------------------------------------------------------\\
# parse_section: Create a hash from a single section of a dhcpd.conf.master file
#
# Parameters: The raw contents of the section as read from the file (array ref)
#             The type of this section (2 = subnet, 3 = pool, 4 = class, 5 = shared)
#             A IP (if this is a subnet)
#             A netmask (if this is a subnet)
#             A name of this class/shared subnet
#
# Returns: A hash reference.
#----------------------------------------------------------------------------//
sub parse_section {
    my @contents = @{$_[0]};
    my ($dummy,$type,$ip,$netmask,$name) = @_;
  
    $type = '0' unless $type;
    my $min = my $max = my $other = '';

    my (@options, @groups, @hosts);
    my $comment; 

    while (my $test = shift(@contents)) {
        chomp ($test);
        next if $test =~ /^\s*$/;

        if ($test =~ /^\s*\#/) {
            if (!$comment) {
                $comment = $test;
                $comment =~ s/^\s*\#//g;
            }
        } elsif ($test =~ /^\s*option\s+(\S+)\s+(.+);$/) {
            my $option = Option->new($1, $2);
            push @options, $option;

        } elsif ($test =~ /^\s*host\s+([a-z0-9\-]+)\.([a-z0-9\-\.]*)\s+\{\s+hardware\s+ethernet\s+([a-f0-9\:]+)\s*;\s+fixed-address\s+([\d\.]+)\s*;\s*\}/i) {
            print STDERR "ERROR: Host found in DHCP config!!\n";

        } elsif ($test =~ /^\s*group\s*{\s*$/) {
            my ($new_group, $new_contents) = parse_section(\@contents);
            @contents = @{$new_contents};
            push @groups, $new_group;
    
        } elsif ($test =~ /^\s*subnet\s*([\d\.]+)\s*netmask\s*([\d\.]+)\s*{\s*$/) {

            my ($new_group, $new_contents) = parse_section(\@contents, 2, $1, $2);
            @contents = @{$new_contents};
            push @groups, $new_group;

        } elsif ($test =~ /^\s*pool\s*{\s*$/) {
            my ($new_group, $new_contents) = parse_section(\@contents,3);
            @contents = @{$new_contents};
            push @groups, $new_group;

        } elsif ($test =~ /^\s*class\s*\"([a-z0-9\-_]+)\"\s*{\s*$/i) {
            my ($new_group, $new_contents) = parse_section(\@contents,4, '','',$1);
            @contents = @{$new_contents};
            push @groups, $new_group;

        } elsif ($test =~ /^\s*shared-network\s*\"([a-z0-9\-_]+)\"\s*{\s*$/i) {
            my ($new_group, $new_contents) = parse_section(\@contents,5, '','',$1);
            @contents = @{$new_contents};
            push @groups, $new_group;

        } elsif ($test =~ /^\s*\}\s*$/) {
            last;

        } else {
            if ($type == 3 && $test =~ /^\s*range\s*([\d\.]+)\s*([\d\.]+)\s*;\s*$/) {
                $min = $1;
                $max = $2;
            }
            $test =~ /^\s*([a-z0-9\-\.]+)\s*(.*)\;$/;
            if ($1 && $dhcp_options{$1}) {
                my $option = Option->new($1, $2, 1);
                push @options, $option;
            } else {
                $other .= "\n" if $other;
                $test =~ s/^\s*//;
                $other .= $test;
            }
        }
    }
    my $group;
    if ($type == 2) {
        $group = SubnetDec->new($ip, $netmask);
    } elsif ($type == 3) {
        $group = Pool->new($min, $max);
    } elsif ($type == 4) {
        $group = Class->new($name);
    } elsif ($type == 5) {
        $group = SharedNet->new($name);
    } else {
        $group = Group->new($comment);
    }

    foreach (@options) {
        $group->add_option($_);
    }
    foreach (@groups) {
        $group->add_group($_);
    }

    $group->update_text($other);
    return ($group, \@contents);
}

#----------------------------------------------------------------------------\\
# get_section: Returns the hash associated with the given section/group
#
# Parameters: A section/group id in the form of section:group:group
#             A hash as returned by parse_conf
#
# Returns: A hash reference.
#----------------------------------------------------------------------------//
sub get_section {
    my $section = shift;
    my $hash = shift;

    my @display_group = split(':', $section);
    my $display_hash = $hash;
    foreach (@display_group) {
        $display_hash = $display_hash->get_group($_);
        last if !$display_hash;
    }
    return $display_hash;
}

#----------------------------------------------------------------------------\\
# check_input: Checks the format of user input when adding/modifying a host.
#
# Parameters: None.
#
# Returns: The host's hostname (may be reformatted using pc-assetno
#          A reformatted MAC address.
#----------------------------------------------------------------------------//
sub check_input {
    my %form = get_input();
    my $dynamic = $form{'dynamic'} eq 'true' ? 1 : 0;

    # Format the hostname using pc-assetno if necessary.
    my $hostname;
    if ($form{'id'} || $form{'nonasset'} =~ /true/) {
        $hostname = $form{'id'};
    } else {
        my $serial;
        $serial = db_next_serial($form{'asset'}, $form{'mac'});

        $hostname = $form{'platform'} =~ /printer/i ? 'pr' : 'pc';
        $hostname .= $form{'asset'};
        $hostname .= "-".$serial if $serial;
    }
    my $fqdn = $hostname.".".$form{'domain'};

    # Check IP details
    if (!$dynamic && $form{'ip'} !~ /\d+\.\d+\.\d+\.\d+/) {
        admin_error(pop_up("Invalid IP given - $form{'ip'}"), "Input");
        return 0;
    }
    if (!$dynamic && !what_is_my_registered_subnet($form{'ip'})) {
        admin_error(pop_up("IP is not in a known subnet"), "Input");
        return 0;
    }

    # Check hostname format
    if ($hostname !~ /^[a-z0-9\-]+$/i) {
        admin_error(pop_up("Invalid fqdn given. $fqdn"), "Input");
        return "0";
    }

    # Check hostname privileges
    if ($hostname !~ /^p[cr]\d+[\-\d]*$/ && !priv('custom_hostname')) {
        admin_error(pop_up("Insufficient privileges to specify a custom hostname."), "Input");
        return 0;
    }

    # Ensure the given hostname is unique in both the
    # database AND dns (this call does both).
    $form{'ip'} = '' if $dynamic;
    if (!$form{'update'} && !db_unique_host($hostname, $form{'domain'}, $form{'ip'})) {
        my $dns_ip = gethostbyname($fqdn);
        my $conv_ip = 'Unknown';
        $conv_ip = inet_ntoa($dns_ip) if $dns_ip;
        admin_error(pop_up("Host $fqdn is in use by $conv_ip"),"Input");
        return 0;
    }

    # Make sure we have a VALID domain
    my @domains = @{read_domains()};
    my $found = 0;
    foreach (@domains) {
        if ($_ eq $form{'domain'}) {
            $found = 1;
            last;
        }
    }
    if (!$found) {
        admin_error(pop_up("Domain not found in the list of available domains."), "Input");
        return "0";
    }

    # Check assest numbers
    if ($form{'nonasset'} !~ /true/ && $form{'asset'} !~ /^\d\d\d\d\d\d$/) {
        admin_error(pop_up("Asset Number must be 6 digits exactly. ".
            "Please prefix it with Zeros if it's a shorter number."), "Input");
        return 0;
    }
    if ($form{'nonasset'} =~ /true/ && !priv('non_gu_asset')) {
        register_error("Only superusers may register non-gu assets.");
    }

    # Check staff/student numbers
    if ($form{'owner'} !~ /^s\d+$/i && !priv('relaxed_s_number')) {
        admin_error(pop_up("Owner must be 's' followed by numbers."), "Input");
        return 0;
    }

    # Check mac format
    if ($form{'mac'} !~ /^([0-9a-f]{1,2}).??([0-9a-f]{1,2}).??([0-9a-f]{1,2}).??([0-9a-f]{1,2}).??([0-9a-f]{1,2}).??([0-9a-f]{1,2})$/i) {
        admin_error(pop_up("MAC must be in xx:xx:xx:xx:xx:xx format, ".
                           "where 'x' = a hex digit."), "Input");
        return 0;
    }
    $form{'mac'} = uc sprintf("%02s:%02s:%02s:%02s:%02s:%02s",$1,$2,$3,$4,$5,$6);

    my @row = db_get_all($form{'mac'});
    if (!$form{'update'} && $row[0]) {
        admin_error(pop_up("This MAC is already registered with Netreg"), "Input");
        return 0;
    }

    # Check expiry date
    if ($form{'expiry'} && $form{'expiry'} !~ /^\d\d\/\d\d\/\d\d\d\d$/) {
        admin_error(pop_up("Expiry date not in the format dd/mm/yyyy"),"Input");
    }

    # Check MX records
    if ($form{'mx'}) {
        my @records = split(',', $form{'mx'});
        foreach (@records) {
            if (!/^(\d+)\s+([a-z\-0-9\.]+)$/) {
                admin_error(pop_up("Invalid MX record format."),"Input");
                return 0;
            }
        }
    }

    # Check CNAME records
    if ($form{'cname'}) {
        my @records = split(',', $form{'cname'});
        foreach (@records) {
            if (!/^([a-z\-0-9\.]+)$/) {
                admin_error(pop_up("Invalid CNAME record format."),"Input");
                return 0;
            }

            # Make sure we have a valid domain.
            my $found = 0;
            push @domains, @ZONES;
            foreach (@domains) {
                if ($form{'cname'} =~ /^[0-9a-z\-]+\.$_/) {
                    $found = 1;
                    last;
                }
            }
            if (!$found) {
                admin_error(pop_up("CNAME domain is not valid."), "Input");
                return "0";
            }
        }
    }
    return ($hostname, $form{'mac'});
}

#----------------------------------------------------------------------------\\
# pop_up: Returns a string of javascript designed to pop-up the host-entry
# window in the event of an error so the details may be re-entered.
#
# Parameters: An error message.
#
# Returns: A string containing the error message and javascript to be printed.
#----------------------------------------------------------------------------//
sub pop_up {
    my $error = shift;
    my %form = get_input();

    my $values = '';
    foreach my $key (keys %form) {
        next if $key eq 'action';
        $values .= "&$key=".$form{$key};
    }
    my $ret = "<script language=javascript>\n".
              "//<!--\n".
              "    window.open('$ADMINPATH/admin.cgi?action=statichostretry$values&error=$error',\n".
              "              'Confirm_Action',\n".
              "              'toolbar=no,location=no,directories=no,status=no,menubar=no,scrollbars=yes,' +\n".
              "              'resizable=yes,width=550,height=450');\n".
              "//-->\n".
              "</script>\n";
    return "$error\n$ret";
}

#----------------------------------------------------------------------------\\
# reverse_ip: Converts an ip to "x.x.x.x.in-addr.arpa" form.
#
# Parameters: An ip address (x.x.x.x).
#
# Returns: The reversed IP in x.x.x.x.in-addr.arpa form.
#----------------------------------------------------------------------------//
sub reverse_ip {
    my $ip = shift;
    $ip =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/;
    return "$4.$3.$2.$1.IN-ADDR.ARPA";
}
#----------------------------------------------------------------------------/
1;
