#----------------------------------------------------------------------------\
# NetReg 2.0 (pre-release devel version)
#       DHCPD Registration Frontend
#
# Version 1.3: 
# Copyright (C) 1999-2002 Peter Valian, Southwestern University
# Version 2.0:
# Portions Copyright (C) 1999-2002 Peter Valian, Southwestern University.
# Portions Copyright (C) 2002 David Bussenschutt, Griffith University, Australia.
# Portions Copyright (C) 2002 Robert Lowe,  Lawrence University, Wisconsin.
# Portions Copyright (C) 2002-2003 Michael McMaster, Griffith University, Australia.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Version 1.3: http://www.netreg.org/
# Version 2.0: http://www.sourceforge.net/projects/netreg/
#
# Version 1.3 code contributions by:
#   Peter Valian <valianp@southwestern.edu>,  Southwestern University.
#     Author of Version 1.3rc2 
#   Bobby Clark <bobby.clark@eku.edu>, Eastern Kentucky University 
#     IMAP Authentication
#   Michael Grinnell <grinnell@american.edu>, American University
#     LDAP Authentication
#
# Version 2.0 code contributions by:
#   Robert Lowe <robert.h.lowe@lawrence.edu>, Lawrence University, Wisconsin.
#     CIDR (Subnet Masks etc) Handling/Notation
#     Alpha Testing, and bug chasing in David's Code.
#   David Bussenschutt <D.Bussenschutt@mailbox.gu.edu.au>, Griffith University, Queensland, Australia.
#     Much re-arrangement, and fixes, including but not limited to:
#     Perl 5 Syntax, and modules
#     File Locking
#     OMAPI Support
#     Remote dhcpdmanager daemon, and client support (functionally similar to omapi, but more fully implemented)
#     Improved Error Handling
#     Use of -w for improved code
#     LDAP group handling, for Permissions.
#     install_script
#   Michael McMaster <m4k3r@users.sourceforge.net>, Griffith University,  Queensland, Australia.
#     DBI/Oracle Support
#     HTML extraction/templating
#     Permissions handling (access controls)
#     Cookie based authentication
#     Interface for editing dhcp options/groups and hosts.
#     Synchronisation between hosts
#     Scalability improvements
#     Various code clean-ups
#     Consistency checker for host information 
#----------------------------------------------------------------------------/
package SubnetDec;
use strict;
use warnings;
our @ISA = ( "Group" );

#----------------------------------------------------------------------------\\
#
# Imported Modules
#
#----------------------------------------------------------------------------//
use Group qw (write_hosts);
use Html qw (print_html_file);

#----------------------------------------------------------------------------\\
#
# Constants
#
#----------------------------------------------------------------------------//
# A hash matching masks in 255.255.255.0 format to CIDR notation.
my %masks;

#----------------------------------------------------------------------------\\
#
# Automatically Executed Methods
#
#----------------------------------------------------------------------------//
BEGIN {
    my $max = 2**32;
    for (my $i=0; $i<32; $i++) {
         $masks{$max - 2**$i} = 32 - $i;
    }
}

#----------------------------------------------------------------------------\\
# new: Creates a new "SubnetDec" object. (Stores info about Subnet declarations)
#
# Parameters: The ip and netmask of this subnet. eg. subnet 10.0.0.0 netmask 255...
#
# Returns: A SubnetDec object (a blessed hash reference).
#----------------------------------------------------------------------------//
sub new {
    my $class = shift;
    my $ip = shift;
    my $netmask = shift;
    my $self  = {}; 

    $self->{'options'} = {};
    $self->{'hosts'} = {};
    $self->{'groups'} = {};
    $self->{'text'} = '';

    $self->{'ip'} = $ip;
    $self->{'netmask'} = $netmask;

    bless($self, $class);    
    return $self;
}   

#----------------------------------------------------------------------------\\
# write: write the contents of this subnet to a given file handle in dhcpd.conf
# format.
#
# Parameters: A filehandle to write to.
#             An indentation level
#             The cumulative URL of this group. (eg. g1:g2:g3...)
#             True if hosts from the database should be inserted.
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub write {
    my $self = shift;
    my $fh = shift;
    my $level = shift || 0;
    my $c_url = shift;
    my $hosts = shift;

    $c_url .= ':' if $c_url;
    $self->{'netmask'} =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/;
    $c_url .=  $self->{'ip'}.'/'.$masks{($1 << 24) + ($2 << 16) + ($3 << 8) + $4};

    pad ($fh, $level - 1);
    print $fh "subnet ".$self->{'ip'}." netmask ".$self->{'netmask'}." {\n";

    my @lines = split ('\n', $self->{'text'});
    foreach (@lines) {
        pad ($fh, $level); 
        print $fh "$_\n";
    }
    foreach (sort keys %{$self->{'options'}}) {
        pad ($fh, $level);
        $self->{'options'}{$_}->write($fh);
    }
 
    write_hosts($fh,$c_url,$level) if $hosts;

    foreach (sort keys %{$self->{'groups'}}) {
        $self->{'groups'}{$_}->write($fh, $level+1,$c_url,$hosts);
    }
 
    print $fh "\n";
    pad ($fh, $level - 1);
    print $fh "}\n";
}

#----------------------------------------------------------------------------\\
# name: returns the name of this subnet.
#
# Parameters: None.
#
# Returns: The name of this subnet (ip/mask).
#----------------------------------------------------------------------------//
sub name {
    my $self = shift;

    $self->{'netmask'} =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/;
    return $self->{'ip'}.'/'.$masks{($1 << 24) + ($2 << 16) + ($3 << 8) + $4};
}

#----------------------------------------------------------------------------\\
#
# Non-Object Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# pad: prints a given number of spaces to a filehandle.
#
# Parameters: A filehandle to print to.
#             The level of padding (4 times this number will be printed)
#
# Returns: Nothing useful.
#----------------------------------------------------------------------------//
sub pad {
    my $fh = shift;
    my $level = 4 * shift;
    print $fh sprintf("%$level".'s','') if $level;
}
1;
