#----------------------------------------------------------------------------\
# NetReg 2.0 (pre-release devel version)
#       DHCPD Registration Frontend
#
# Version 1.3: 
# Copyright (C) 1999-2002 Peter Valian, Southwestern University
# Version 2.0:
# Portions Copyright (C) 1999-2002 Peter Valian, Southwestern University.
# Portions Copyright (C) 2002 David Bussenschutt, Griffith University, Australia.
# Portions Copyright (C) 2002 Robert Lowe,  Lawrence University, Wisconsin.
# Portions Copyright (C) 2002-2003 Michael McMaster, Griffith University, Australia.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Version 1.3: http://www.netreg.org/
# Version 2.0: http://www.sourceforge.net/projects/netreg/
#
# Version 1.3 code contributions by:
#   Peter Valian <valianp@southwestern.edu>,  Southwestern University.
#     Author of Version 1.3rc2 
#   Bobby Clark <bobby.clark@eku.edu>, Eastern Kentucky University 
#     IMAP Authentication
#   Michael Grinnell <grinnell@american.edu>, American University
#     LDAP Authentication
#
# Version 2.0 code contributions by:
#   Robert Lowe <robert.h.lowe@lawrence.edu>, Lawrence University, Wisconsin.
#     CIDR (Subnet Masks etc) Handling/Notation
#     Alpha Testing, and bug chasing in David's Code.
#   David Bussenschutt <D.Bussenschutt@mailbox.gu.edu.au>, Griffith University, Queensland, Australia.
#     Much re-arrangement, and fixes, including but not limited to:
#     Perl 5 Syntax, and modules
#     File Locking
#     OMAPI Support
#     Remote dhcpdmanager daemon, and client support (functionally similar to omapi, but more fully implemented)
#     Improved Error Handling
#     Use of -w for improved code
#     LDAP group handling, for Permissions.
#     install_script
#   Michael McMaster <m4k3r@users.sourceforge.net>, Griffith University,  Queensland, Australia.
#     DBI/Oracle Support
#     HTML extraction/templating
#     Permissions handling (access controls)
#     Cookie based authentication
#     Interface for editing dhcp options/groups and hosts.
#     Synchronisation between hosts
#     Scalability improvements
#     Various code clean-ups
#     Consistency checker for host information 
#----------------------------------------------------------------------------/
# All Real code here by Robert Lowe.
# perl-5isms, locking, _urange  and what_ sub/s, and tweaks by David Bussenschutt.
package Subnets;
use strict;
use warnings;
#----------------------------------------------------------------------------\\
#
# Exports
#
#----------------------------------------------------------------------------//
use Exporter();
our @ISA = qw(Exporter);

our @EXPORT = qw( );
our @EXPORT_OK = qw ( ip_in_subnet ip_in_subnet_urange ip2dec read_domains
                      what_is_my_registered_subnet db_init dec2ip ip_in_subnet_d
                      ip_in_subnet_urange_d ipv4_cidr2msk add_new_subnet_to_system );
#----------------------------------------------------------------------------\\
#
# Imported Modules
#
#----------------------------------------------------------------------------//
# If using mod_perl, the $FindBin::RealBin directory will be set ONCE only
# see startup.pl
use FindBin;
use lib "$FindBin::RealBin"; # locate library path for below modules.

# Netreg modules
use Html qw(admin_error);
use Variables qw( $DHCPDMASTERLOCK $STATIC_FILE $GREP $GLOBAL_AUTH_SVR $GLOBAL_AUTH_METHOD );
use Database qw(db_connect);
use NetregMisc qw(open_with_lock append_error_log);

#----------------------------------------------------------------------------\\
#
# Constants
#
#----------------------------------------------------------------------------//
# Optional mod_perl support.
# Sets a constant, and ensures the correct exit() function is called.
use constant IS_MODPERL => $ENV{MOD_PERL};
use subs qw(exit);
use POSIX qw(floor);
# Select the correct exit function
*exit = IS_MODPERL ? \&Apache::exit : sub { CORE::exit };

#----------------------------------------------------------------------------\\
#
# Global Variables
#
# These are shared within this perl module only.
# These variables are specifically shared between executions if using mod_perl
#----------------------------------------------------------------------------//
my %masks;              # A hash of all possible subnet masks.
my %u_range;            # A hash of subnets to an array ref of unreg ip range.

#----------------------------------------------------------------------------\\
#
# Automatically Executed Methods
#
#----------------------------------------------------------------------------//
BEGIN {
    my $max = 2**32;
    for (my $i=0; $i<32; $i++) {
        $masks{32-$i} = $max - 2**$i;
    }
}                      

#----------------------------------------------------------------------------\\
#
# SQL init
#
#----------------------------------------------------------------------------//
# Prepare database queries. This function is only called by db_connect in
# Database.pm

my ($sql_get_domains,$sql_get_all_domains,$sql_get_ranges,$sql_get_subnets);
sub db_init {
    my $dbh = shift;

    $sql_get_domains = $dbh->prepare("select domain from subnet_domain where ".
        "cidr_subnet = ? and cidr_mask = ?");
    $sql_get_all_domains = $dbh->prepare("select domain from subnet_domain");
    $sql_get_ranges = $dbh->prepare("select * from ip_range where ".
        "cidr_subnet = ? and cidr_mask = ?");
    $sql_get_subnets = $dbh->prepare("select * from subnets");
}

#----------------------------------------------------------------------------\\
#
# Exported Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# read_domains: Returns a list of valid domains for a given IP
#
# Parameters: An IP to find valid domains for, or undef for all domains
#
# Returns: An array reference of domains or undef on error.
#----------------------------------------------------------------------------//
sub read_domains {
    my $tmpIP = shift || '';

    my ($subnet, $mask) = what_is_my_registered_subnet($tmpIP);
    
    my @domains;
    my $sql_executed;
    if ($subnet) { 
        $sql_get_domains->execute($subnet, $mask);
        $sql_executed = $sql_get_domains;
    } else {
        $sql_get_all_domains->execute();
        $sql_executed = $sql_get_all_domains;
    }
  
    while (my @row = $sql_executed->fetchrow_array()) {
        push @domains, $row[0];
    }
    $sql_executed->finish();        

    # Remove any duplicates. This code is taken from the Perl Cookbook.
    my %seen = ();
    my @unique = grep { ! $seen{$_} ++ } @domains;

    return \@unique;
}

#----------------------------------------------------------------------------\\
# ip_in_subnet: Test whether an ip is within a given subnet.
#
# Parameters: IP to test
#             subnet/mask
#
# Returns: True if IP in subnet, or else false
#----------------------------------------------------------------------------//
sub ip_in_subnet {
    my $ip = shift;
    my $subnet = shift;
    my $mask = shift;

    my $pip = ip2dec($ip);
    my $psn = ip2dec($subnet);
    
    return ($pip & $masks{$mask}) == ($psn & $masks{$mask});
}

#----------------------------------------------------------------------------\\
# ip_in_subnet_d: Test whether an ip is within a given subnet.
#
# Parameters: IP to test, in decimal format
#             subnet(decimal)/mask
#
# Returns: True if IP in subnet, or else false
#----------------------------------------------------------------------------//
sub ip_in_subnet_d {
    my $ip = shift;
    my $subnet = shift;
    my $mask = shift;

    return ($ip & $masks{$mask}) == ($subnet & $masks{$mask});
}

#----------------------------------------------------------------------------\\
# ip_in_subnet_urange: Test whether an ip is within a given subnet's unregistered
# range.
#
# Parameters: 
#             The IP to test
#             Subnet/mask
#
# Returns: True if IP in the unreg range, or else false
#----------------------------------------------------------------------------//
sub ip_in_subnet_urange {
    my $ip = shift;
    my $subnet = shift;
    my $mask = shift;

    if (!$u_range{"$subnet/$mask"}) {
        my @array = ();
        $sql_get_ranges->execute($subnet, $mask);
        while (my @row = $sql_get_ranges->fetchrow_array) {
           push @array, ip2dec($row[0])."#".ip2dec($row[1]);
        }
        $sql_get_ranges->finish();
        $u_range{"$subnet/$mask"} = \@array;
    }

    my $IPint = ip2dec( $ip );
    my $found;
    foreach (@{$u_range{"$subnet/$mask"}}) {
        my ($lo,$hi) = split ('#',$_);
        return 1 if ($IPint >= $lo  && $IPint <= $hi );
    }
    return 0;
}   

#----------------------------------------------------------------------------\\
# ip_in_subnet_urange_d: Test whether an ip is within a given subnet's 
# unregistered range.
#
# Parameters: The IP to test in decimal format
#             Subnet/mask
#
# Returns: True if IP in the unreg range, or else false
#----------------------------------------------------------------------------//
sub ip_in_subnet_urange_d {
    my $ip = shift;
    my $subnet = shift;
    my $mask = shift;

    if (!$u_range{"$subnet/$mask"}) {
        my @array = ();
        $sql_get_ranges->execute($subnet, $mask);
        while (my @row = $sql_get_ranges->fetchrow_array) {
           push @array, ip2dec($row[0])."#".ip2dec($row[1]);
        }
        $sql_get_ranges->finish();
        $u_range{"$subnet/$mask"} = \@array;
    }

    my $found;
    foreach (@{$u_range{"$subnet/$mask"}}) {
        my ($lo,$hi) = split ('#',$_); 
        return 1 if ($ip >= $lo  && $ip <= $hi );
    }
    return 0;
}

#----------------------------------------------------------------------------\\
# what_is_my_registered_subnet: Search for the subnet a given ip exists in.
#
# Parameters: 
#             The IP to find a subnet for
#
# Returns: A subnet, or undef if none found.
#----------------------------------------------------------------------------//
sub what_is_my_registered_subnet {
    my $ip = shift; 

    return undef unless $ip =~ /^\d+.\d+.\d+.\d+$/;

    $sql_get_subnets->execute();
    while (my @row = $sql_get_subnets->fetchrow_array) {
    # check each reg and unreg subnet in turn until we get one that tells us that the IP is valid there.
        if (ip_in_subnet($ip,$row[0],$row[1])) {
            $sql_get_subnets->finish();
            return ($row[0],$row[1]);
        }
    }
    $sql_get_subnets->finish();
    return undef;
}

#----------------------------------------------------------------------------\\
# add_new_subnet_to_system: this function adds a new subnet declaration to the 
# dhcpd.conf.master file AND it adds the required settings into the database 
# for the new subnet to be used. (this used to be in add_subnet.pl)
#
# parameters: section - the full name of the section in the dhcpd.conf.master file to 
#			put this subnet declaration inside. eg: "NonRoaming:Blah:Blah2"
#	      campus - must be a valid campus from the database
#	      cidrmask - must be a valid mask (a num from 1-31)
#	      network - a string like '123.123.123.0' would be correct for a cidrmask of 24
#	      description - a meaningful string that is displayed on the subnet overview page.
#	      domains - a list of domains that are valid for that subnet.
#
#----------------------------------------------------------------------------//
sub add_new_subnet_to_system {
my ($section,$campus,$cidrmask,$network,$description,@domains) = @_;

my $dbh = Database::db_connect(1) || die "Could not connect to database.\n";

# convert the CIDR mask into the traditional xxx.xxx.xxx.xxx netmask:
my $netmask = ipv4_cidr2msk($cidrmask);                                                                                                                                              
my $sql_servers = $dbh->prepare("select server_ip from campuses where campus = ?");
$sql_servers->execute($campus);
my $servers = '';
while (my @server_row = $sql_servers->fetchrow_array) {
    $servers .= "," if $servers;
    $servers .= " $server_row[0] ";
}
$sql_servers->finish();
                                                                                                                                              
# Lock the entire dhcpd_master file here to ensure corruption
# Cannot occur.
my $lock_file = open_with_lock("> $DHCPDMASTERLOCK");
if (!$lock_file) {
    append_error_log("Could not lock file $DHCPDMASTERLOCK\n");
}
                                                                                                                                              
my $out_lines = '';
open (CONF, "< $STATIC_FILE") || do {
    close $lock_file;
    append_error_log("Could not open $STATIC_FILE\n");
};
                                                                                                                                              
# get my current IP as the DODGY DNS server to use...
my $ip = `/usr/sbin/ifconfig -a | $GREP inet | $GREP -v 127.0.0.1`;
chomp $ip;
my @ip = split ("\n", $ip);
# choose the first ip you find...
foreach (@ip) {
    s/mask.*//i; # remove broadcast and netmask IP info
    s/cast.*//i; # remove broadcast and netmask IP info
    $ip=$_,last if s/^.*[\s:](\d+\.\d+\.\d+\.\d+)\s+.*$/$1/;
}
                                                                                                                                              
# HACK, we currently locate the "group" in the dhcpd.conf.master
# file by just doing a pattern match on the last section/element
# of the group heirarchy, which MAY not necessarily be 
# unique, but should be in most cases.
$section =~ s/^.*://g;  # strip off the leading path to the group..


#------START OF create a few variables for populating into the config below---------
$network =~ /^\d+\.\d+\.(\d+)\.\d+$/;
my $network_bit = "132.234.$1";

# DEFAULT SETTINGS:
# TODO: ask the user to tell us the following, rather than
# assume a /24 network with a range/setup like Griffith University...
my $router = $network_bit.".209";
my $reg_range_lower = $network_bit.".20";
my $reg_range_upper = $network_bit.".200";
my $ureg_range_lower = $network_bit.".220";
my $ureg_range_upper = $network_bit.".250";
#-------END OF create a few variables for populating into the config below--------


my $inserted = 0; 
while (my $line = <CONF>) {
    chomp($line);
    $out_lines .= $line."\n";

    if (($line =~ /^\s*\#\s*$section\s*$/i) && $inserted == 0) {
	$inserted++;	

#---------------------NEW SUBNET DECLARATION LAYOUT START-----------------
        $out_lines .= "subnet $network netmask $netmask {\n".
                      "authoritative;\n".
                      "option routers $router;\n".
                      "default-lease-time 28800;\n".
                      "option domain-name-servers $servers;\n".
                      "pool {\n".
                          "failover peer \"dhcp\";\n".
                          "deny dynamic bootp clients;\n".
                          "range $reg_range_lower $reg_range_upper;\n".
                          "deny unknown clients;\n".
                      "}\n".
                      "pool {\n".
                          "failover peer \"dhcp\";\n".
                          "deny dynamic bootp clients;\n".
                          "allow unknown clients;\n".
                          "default-lease-time 240;\n".
                          "max-lease-time 240;\n".
                          "range $ureg_range_lower $ureg_range_upper;\n".
                          "option domain-name-servers $ip;\n".
                      "}\n".
                      "\n}\n";
#---------------------NEW SUBNET DECLARATION LAYOUT END-------------------

    }
}
close CONF;
open (CONF, "> $STATIC_FILE") || do {
    close $lock_file;
    append_error_log("Could not open $STATIC_FILE\n");
};
                                                                                                                                              
print CONF "$out_lines";
close CONF;
                                                                                                                                              
my $sql_subnet = $dbh->prepare("insert into subnets values (?,?,?,?)");
$sql_subnet->execute($network, $cidrmask,$campus,$description);
$sql_subnet->finish();
                                                                                                                                              
my $sql_range = $dbh->prepare("insert into ip_range values (?,?,?,?,?)");
$sql_range->execute($ureg_range_lower,$ureg_range_upper,$network,$cidrmask,'Y');
$sql_range->finish();
                                                                                                                                              
my $sql_auth = $dbh->prepare("insert into authentication values (?,?,?,?)");
$sql_auth->execute($network,$cidrmask,$GLOBAL_AUTH_SVR, $GLOBAL_AUTH_METHOD);
$sql_auth->finish();
                                                                                                                                              
my $sql_domain = $dbh->prepare("insert into subnet_domain values (?,?,?)");
foreach (@domains) {
    $sql_domain->execute($network,$cidrmask,$_);
}
$sql_domain->finish();
                                                                                                                                              
close $lock_file;

}


#----------------------------------------------------------------------------\\
#
# Non-Exported Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# bad_subnet: Test if subnet/mask pair is valid
#
# Parameters: Subnet/mask
#
# Returns: null if fine, error message if error.
#----------------------------------------------------------------------------//
sub bad_subnet {
    my ($subnet,$mask) = split('/', $_[0]);
    my $psn = ip2dec($subnet);

    if (($psn & $masks{$mask}) != $psn) {
        return "Bad subnet $subnet for provided mask /$mask: should be ".
               dec2ip($psn & $masks{$mask});
    }
    return undef;
}

#----------------------------------------------------------------------------\\
# ip2dec: Convert dotted-quad IP address to 32-bit int
#
# Parameters: "xx.xx.xx.xx" style ip.
#
# Returns: decimal encoding of the given ip.
#----------------------------------------------------------------------------//
sub ip2dec {
    # NOTE: due to the strange functions of the "sort" function,
    # The passed value must be copied before using it, otherwise
    # the sort function (as used in subnet_overview) sorts the wrong
    # data. 
    my $val = shift;
    $val =~ s/\/.*$//;
    my ($o1,$o2,$o3,$o4) = split(/\./, $val);
    return 0 unless defined $o4;  # this shortcircuts the function in hte case where it had a bad parameter.
    return ($o1 << 24) + ($o2 << 16) + ($o3 << 8) + $o4;
}

#----------------------------------------------------------------------------\\
# dec2ip: Convert 32-bit int to dotted-quad IP address.
#
# Parameters: Decimal encoded ip.
#
# Returns: "xx.xx.xx.xx" style ip.
#----------------------------------------------------------------------------//
sub dec2ip {
    return join('.',unpack('C4', pack("N", $_[0])));
}

#----------------------------------------------------------------------------/

#----------------------------------------------------------------------------/
# ipv4_cidr2msk:   Convert a CIDR notation mask (eg /24) to a dotted notation
#		 netmask (eg 255.255.255.0)
#
# Parameters: digit between 0 and 32
#
# Returns: "xxx.xxx.xxx.xxx" stype netmask 
#
# This function taken 99% from Net::IPv4Addr on May 29 2003:
#---------------
#    IPv4Addr.pm - Perl module to manipulate IPv4 addresses.
#
#    Author: Francis J. Lacoste <francis@Contre.COM>
#
#    Copyright (C) 1999 Francis J. Lacoste, iNsu Innovations Inc.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms as perl itself.
#---------------
#----------------------------------------------------------------------------/

# Transform a CIDR mask length in a netmask
sub ipv4_cidr2msk($) {
  my $cidr = shift;
  return undef if $cidr < 0 or $cidr > 32;

  my $bits = "1" x $cidr . "0" x (32 - $cidr);

  return join ".", (unpack 'CCCC', pack("B*", $bits ));
}

#----------------------------------------------------------------------------/
1;

