#!/usr/bin/perl -w
#----------------------------------------------------------------------------\
# NetReg 2.0 (pre-release devel version)
#       DHCPD Registration Frontend
#
# Version 1.3: 
# Copyright (C) 1999-2002 Peter Valian, Southwestern University
# Version 2.0:
# Portions Copyright (C) 1999-2002 Peter Valian, Southwestern University.
# Portions Copyright (C) 2002 David Bussenschutt, Griffith University, Australia.
# Portions Copyright (C) 2002 Robert Lowe,  Lawrence University, Wisconsin.
# Portions Copyright (C) 2002-2003 Michael McMaster, Griffith University, Australia.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Version 1.3: http://www.netreg.org/
# Version 2.0: http://www.sourceforge.net/projects/netreg/
#
# Version 1.3 code contributions by:
#   Peter Valian <valianp@southwestern.edu>,  Southwestern University.
#     Author of Version 1.3rc2 
#   Bobby Clark <bobby.clark@eku.edu>, Eastern Kentucky University 
#     IMAP Authentication
#   Michael Grinnell <grinnell@american.edu>, American University
#     LDAP Authentication
#
# Version 2.0 code contributions by:
#   Robert Lowe <robert.h.lowe@lawrence.edu>, Lawrence University, Wisconsin.
#     CIDR (Subnet Masks etc) Handling/Notation
#     Alpha Testing, and bug chasing in David's Code.
#   David Bussenschutt <D.Bussenschutt@mailbox.gu.edu.au>, Griffith University, Queensland, Australia.
#     Much re-arrangement, and fixes, including but not limited to:
#     Perl 5 Syntax, and modules
#     File Locking
#     OMAPI Support
#     Remote dhcpdmanager daemon, and client support (functionally similar to omapi, but more fully implemented)
#     Improved Error Handling
#     Use of -w for improved code
#     LDAP group handling, for Permissions.
#     install_script
#   Michael McMaster <m4k3r@users.sourceforge.net>, Griffith University,  Queensland, Australia.
#     DBI/Oracle Support
#     HTML extraction/templating
#     Permissions handling (access controls)
#     Cookie based authentication
#     Interface for editing dhcp options/groups and hosts.
#     Synchronisation between hosts
#     Scalability improvements
#     Various code clean-ups
#     Consistency checker for host information 
#----------------------------------------------------------------------------/
use strict;
use warnings;

#----------------------------------------------------------------------------\\
#
# Imported Modules
#
#----------------------------------------------------------------------------//

# For timelocal (calculating expiration dates)
use Time::Local;

# Netreg modules.
use Admin         qw (server_manager find_conf find_lease subnet_overview
                      view_subnet display_popup expire_host display_history);

use Subnets       qw( read_domains dec2ip);

use Variables     qw( variables_init $LDAP_URL $ADMINPATH $GFX $IP 
                      $DHCPDMASTERLOCK $SYNC_SCRIPT $DIG @ZONES $STATIC_FILE $DEBUG );

use Html          qw( get_input html_init admin_error get_cookies 
                      print_html_file);

use Authenticate  qw( auth_init user priv logon is_valid_cookie logoff);

use Database      qw( get_dbh db_connect db_get_unregip
                      db_get_all get_subnet_desc db_get_campuses);

use StaticHosts   qw( print_static_page update_static add_static_parameter 
                      delete_static get_cnames get_groups get_js_groups 
		      get_group_indexes get_top_groups);

use NetregMisc    qw( open_with_lock make_time append_error_log );

#----------------------------------------------------------------------------\\
#
# Constants
#
#----------------------------------------------------------------------------//
# Optional mod_perl support.
# Sets a constant, and ensures the correct exit() function is called.
use constant IS_MODPERL => $ENV{MOD_PERL};
use subs qw(exit);
# Select the correct exit function
*exit = IS_MODPERL ? \&Apache::exit : sub { CORE::exit };

#----------------------------------------------------------------------------\\
#
# Main Function
#
#----------------------------------------------------------------------------//

# ensure large pages (find_conf with > 2000 entries) display immediately.
autoflush STDOUT, 1;
autoflush STDERR, 1;

my $dbh = db_connect();
if (!$dbh) {
    admin_error("Could not connect to database.\n");
    print_html_file("admin_footer.html");
}

html_init();
variables_init();
auth_init();

# %form is NOT shared by mod_perl, because this script will be wrapped and
# it won't be in the outer scope anymore.
my %form = get_input();
do_logon();

# Need to modify IP for hook into IPAudit
# IPAudit likes IPs in the form of xxx.xxx.xxx.xxx
if (defined $form{'src'} && $form{'src'} eq "ipaudit") {
    my @octects = split /\./,$form{'query'};
    # Cast each string of numbers into an integer -- gotta love Perl ;)
    foreach my $octet (@octects) {
        $octet = $octet - 0;
    }
    $form{'query'} = join(".",@octects);
}

# default page..
my $action = $form{'action'};
if (!$action) {
    server_manager();
}
elsif ($action eq "SM")                   { server_manager();   }
elsif ($action eq "FC")                   { find_conf();        }
elsif ($action eq "FL")                   { find_lease();       }
elsif ($action eq "SO")                   { subnet_overview();  }
elsif ($action eq "VS")                   { view_subnet();      }
elsif ($action eq "LO")                   { do_logoff();        }
elsif ($action eq "POP-UP")               { display_popup();    }
elsif ($action eq "RM")                   { expire_host();      }
elsif ($action eq "HISTORY")              { display_history();  }
elsif ($action eq "view-static")          { view_static_hosts();}
elsif ($action eq "update-static")        { deal_with_static(); }
elsif ($action eq "add-static-parameter") { deal_with_static(); }
elsif ($action eq "force-update")         { force_update();     }
elsif ($action eq "statichostadd")        { deal_with_static(); }
elsif ($action eq "statichostretry")      { deal_with_static(); }
elsif ($action eq "statichostconfirm")    { deal_with_static(); }
elsif ($action eq "subnetadd")             { deal_with_static(); }

else { admin_error("$action: unknown action"); }
exit;

#----------------------------------------------------------------------------\\
#
# Methods
#
#----------------------------------------------------------------------------//

#----------------------------------------------------------------------------\\
# do_logoff: logs a user off the system, and displays the logon page.
#
# Parameters: None.
#
# Returns: Nothing.
#----------------------------------------------------------------------------//
sub do_logoff {
    logoff(user());
    my $image = "background-image: url(\"/gfx/logo_6.jpg\");\n".
                "background-repeat: no-repeat;\n".
                "background-position: center center;\n";
    print_html_file('admin_header.html', "dummy;", $image);
    print_html_file('admin_logon.html');
    print_html_file('admin_footer.html');
}

#----------------------------------------------------------------------------\\
# deal_with_static: Updates or adds static hosts.
#
# Parameters: None.
#
# Returns: Nothing.
#----------------------------------------------------------------------------// 
sub deal_with_static {
    # Lock the entire dhcpd_master file here to ensure corruption
    # Cannot occur.
    my $lock_file = open_with_lock("> $DHCPDMASTERLOCK"); 
    if (!$lock_file) {
        admin_error("Could not exclusively open $DHCPDMASTERLOCK lock file.");
        print_html_file("admin_footer.html");
    }

    my %form = get_input();
    my $success = '';

    if ($form{'action'} eq 'add-static-parameter') {
        $success = add_static_parameter($lock_file);
    } elsif ($form{'submit'} && $form{'submit'} eq 'Update') {
        $success = update_static($lock_file); 
    } elsif ($form{'submit'} && $form{'submit'} eq 'Delete') {
        $success = delete_static($lock_file);
    } elsif ($form{'action'} eq 'statichostconfirm') {
        close $lock_file;

        my $dnsconfirmmsg = "The following resource records exist:\n";
        my $host = $form{'host'};
        my $output = `$DIG $host ANY`;

        if ($output =~ /\;\; ANSWER SECTION/) {
            my @lines = split ('\n', $output);
            my $in_section = 0;
            foreach (@lines) {
                if (/;; ANSWER SECTION:/ || $in_section) {
                    $in_section++;
                    next if /;; ANSWER SECTION:/;
                    next if /^\s*$/;
                    last if (/^;/);
                    $dnsconfirmmsg .= "$_\n";
                }
            }
        }
        my @cnames = get_cnames($host);
        foreach (@cnames) {
            $dnsconfirmmsg .= "CNAME $_\n";
        }

        print_html_file("static_confirmdelete.html", "\\^SECTION", $form{'url'},
                        "\\^HASH", $form{'hash'}, "\\^HOST", $form{'host'},
                        "\\^MESSAGE", $dnsconfirmmsg);
        exit;

    } elsif ($form{'action'} eq 'statichostretry') {
        close $lock_file;

        my $domains = read_domains();
        my $domain_options = '';
        foreach (sort @$domains) {
            $domain_options .= "<option value=\"$_\">$_ </option>\n";
        }


        my $levels = get_top_groups();
        my $select_groups = '';
        for (my $t = 1; $t<=$levels; $t++ ) {
            $select_groups .= "<SELECT NAME=\"My$t\" onChange=\"\n";
            for( my $y = $t+1; $y<=$levels; $y++) {
              $select_groups .= "fillSelectFromArray(document.forms[0].My$y,array1";
	 	for ( my $z = 1; $z < $y; $z++) {
		  $select_groups .= "[document.forms[0].My$z.selectedIndex+1]";
		}
	      $select_groups .= ");\n";
            }
            #$select_groups .= "\" DISABLED><OPTION></OPTION>\n</SELECT>\n";
            $select_groups .= "\"><OPTION></OPTION>\n</SELECT>\n";
        }

        my @array_groups = get_js_groups();
        my $array_groups = '';
        foreach ( @array_groups ) {
          $array_groups .= $_;
        }

        my @form_groups = split(/:/, $form{'group'});

        my %group_indexes = get_group_indexes();


        my $onload_groups = '';
        for (my $y = 1; $y<=$#form_groups+2; $y++ ) {
          $onload_groups .= "fillSelectFromArray(document.forms[0].My$y,array1";
          my $frm_grp = $form_groups[$y-1] || '';
          my $index = $group_indexes{$frm_grp} || '';

            $group_indexes{''} = $index."[2]"; # remember for the empty group name (last iteration)

          $index =~ s/\[.\]$//;
          $onload_groups .= $index;
          $onload_groups .= ",'".$frm_grp."');\n";
        }

        my $submit_groups = '';
        for ( my $l=1 ; $l <= $levels ; $l++ ) {
         $submit_groups .= "document.forms[0].My".$l."[document.forms[0].My$l.selectedIndex].value + ':' +\n";
        }

	# disable the MAC address field if we are doing a update...
        my $disabled = '';
        $disabled = 'DISABLED' if ($form{'update'});
      
        # disable the host and domain name fields if we are doing an update, and hte user doesn't have permissions.	
        my $namedisabled = '';
	if (!priv("custom_hostname")) {
          $namedisabled = 'DISABLED' if ($form{'update'});
 	}

        my $nonasset = '';
        $nonasset = 'CHECKED' if (!$form{'asset'});
        $form{'asset'} = '' if (!$form{'asset'});

        my $dynamic = '';
        $dynamic = 'CHECKED' if ($form{'dynamic'} =~ /true/);
        $form{'ip'} = '' if $dynamic eq 'CHECKED';
            
        print_html_file("static_addhost.html", "\\^SECTION", $form{'section'},
                        "\\^HASH", $form{'hash'}, "\\^UPDATE", $form{'update'},
                        "\\^MAC", $form{'mac'}, "\\^IP", $form{'ip'}, "\\^PLATFORM", $form{'platform'},
                        "\\^OWNER", $form{'owner'}, "\\^ASSET", $form{'asset'}, "\\^COMMENT", $form{'comment'},
                        "\\^DISABLED", $disabled, "\\^HOST", $form{'id'}, "\\^DOMAIN", $form{'domain'},	
			"\\^NAMEDISABLED", $namedisabled,
                        "\\^OPTIONS", $domain_options, "\\^ERROR", $form{'error'}, 
			"\\^NONASSET", $nonasset, "\\^MX", $form{'mx'}, "\\^CNAME", $form{'cname'},
			"\\^SELECT_GROUPS", $select_groups, "\\^ARRAY_GROUPS", $array_groups,
                        "\\^ONLOAD_GROUPS", $onload_groups, "\\^SUBMIT_GROUPS", $submit_groups,
                        "\\^DYNAMIC", $dynamic, "\\^EXPIRY", $form{'expiry'});
        exit;

    } elsif ($form{'action'} eq 'statichostadd') {
        close $lock_file;

        my ($isUpdate,$mac,$ip,$platform,$owner,$asset,$comment,$disabled,$host,$domain, $nonasset, $domain_options, $mx, $cname,$group_options,$expiry,$dynamic) = '';
        if ($form{'mac'}) {
            $isUpdate++;
            $mac = $form{'mac'};
            my @row = db_get_all($mac);
            ($ip, $owner, $asset, $comment, $host, $domain, $platform,$expiry) = 
                  ($row[9], $row[3], $row[2], $row[8], $row[1], $row[4],$row[7],$row[6]);
            if ($asset =~ /non/i) {
                $asset = '';
                $nonasset = 'CHECKED';
            }

   	    #$disabled = 'DISABLED';

            $dynamic = 'CHECKED' if $row[5] ne 'S'; 
   
            $expiry =~ s/(....)(..)(..)/$3\/$2\/$1/;

            # TODO Get a list of CNAME's
            my @names = get_cnames("$host.$domain");
            $cname = join(',', @names);

            # Get a list of MX records
            my $output = `$DIG $host.$domain MX`;
            if ($output =~ /\;\; ANSWER SECTION/) {
                my @lines = split ('\n', $output);
                my $in_section = 0;
                foreach (@lines) {
                    if (/;; ANSWER SECTION:/ || $in_section) {
                        $in_section++;
                        next if /;; ANSWER SECTION:/;
                        next if /^\s*$/;
                        last if (/^;/);
                        /MX\s*(\d+)\s*(.+)$/;
                        $mx .= ',' if $mx;
                        $mx .= "$1 $2";
                    }
                }
            }
        }


        # disable the MAC address field if we are doing a update...
        $disabled = '';
        $disabled = 'DISABLED' if ($form{'mac'});
                                                                                                                                            
        # disable the host and domain name fields if we are doing an update, and the user doesn't have permissions.
        my $namedisabled = '';
        if (!priv("custom_hostname")) {
          $namedisabled = 'DISABLED' if ($form{'mac'});
        }

         
        my $domains = read_domains();
        foreach (sort @$domains) {
            $domain_options .= "<option value=\"$_\">$_ </option>\n";
        }

        my $levels = get_top_groups();
        my $select_groups = '';
        for (my $t = 1; $t<=$levels; $t++ ) {
            $select_groups .= "<SELECT NAME=\"My$t\" onChange=\"\n";
            for( my $y = $t+1; $y<=$levels; $y++) {
              $select_groups .= "fillSelectFromArray(document.forms[0].My$y,array1";
                for ( my $z = 1; $z < $y; $z++) {
                  $select_groups .= "[document.forms[0].My$z.selectedIndex+1]";
                }
              $select_groups .= ");\n";
            }
            #$select_groups .= "\" DISABLED><OPTION></OPTION>\n</SELECT>\n";
            $select_groups .= "\"><OPTION></OPTION>\n</SELECT>\n";
        }

        my @array_groups = get_js_groups();
	my $array_groups = '';
	foreach ( @array_groups ) {
	  $array_groups .= $_;
	}

        my @form_groups = split(/:/, $form{'url'});

        my %group_indexes = get_group_indexes();

        my $onload_groups = '';
        for (my $y = 1; $y<=$#form_groups+2; $y++ ) {
          $onload_groups .= "fillSelectFromArray(document.forms[0].My$y,array1";
          my $frm_grp = $form_groups[$y-1] || '';
          my $index = $group_indexes{$frm_grp} || '';
             $group_indexes{''} = $index."[2]"; # remember for the empty group name (last iteration)
          $index =~ s/\[.\]$//;
          $onload_groups .= $index;
          $onload_groups .= ",'".$frm_grp."');\n"; 
        }
        my $submit_groups = '';
        for ( my $l=1 ; $l <= $levels ; $l++ ) {
         $submit_groups .= "document.forms[0].My".$l."[document.forms[0].My$l.selectedIndex].value + ':' +\n";
        }

        
        print_html_file("static_addhost.html", "\\^SECTION", $form{'url'},
                        "\\^HASH", $form{'hash'}, "\\^UPDATE", $isUpdate,
                        "\\^MAC", $mac, "\\^IP", $ip, "\\^PLATFORM", $platform,
                        "\\^OWNER", $owner, "\\^ASSET", $asset, "\\^COMMENT", $comment,
                        "\\^DISABLED", $disabled, "\\^HOST", $host, "\\^DOMAIN", $domain,
			 "\\^NAMEDISABLED", $namedisabled,
                        "\\^OPTIONS", $domain_options, "\\^ERROR", "", "\\^NONASSET", $nonasset, 
			"\\^MX", $mx, "\\^CNAME", $cname, 
			"\\^SELECT_GROUPS", $select_groups, "\\^ARRAY_GROUPS", $array_groups,
			"\\^ONLOAD_GROUPS", $onload_groups, "\\^SUBMIT_GROUPS", $submit_groups,
			"\\^EXPIRY", $expiry, "\\^DYNAMIC", $dynamic);

        exit;
    } elsif ($form{'action'} eq 'subnetadd') {

	my @domains = db_get_campuses();
	my $campuses_options = '';
	foreach (sort @domains) {
	  $campuses_options .= "<option value=\"$_\">$_ </option>\n";
	}

	my $domains = read_domains();
        my $domains_options = '';
        foreach (sort @$domains) {
            $domains_options .= "<option value=\"$_\">$_ </option>\n";
        }
                                                                                                                                           
      print_html_file("static_addsubnet.html", 
	          "\\^SECTION", $form{'url'}, "\\^SAVED_MD5", $form{'hash'}, 
		  "\\^CAMPUSES",  $campuses_options,
		  "\\^DOMAINS",   $domains_options,
		   );
       exit ;
    } else {
        admin_error("Unimplemented function.");
        print_html_file("admin_footer.html");
    }    

    close $lock_file;
    if (!$success) {
        admin_error("Update/delete failed. The data may have been modified by ".
                    "someone else and failed the consistency check. ".
                    "Please try again.");
    }
    view_static_hosts();
}

#----------------------------------------------------------------------------\\
# view_static_hosts: Displays the static hosts registered with the system.
# Parameters: None.
#
# Returns: Nothing.
#----------------------------------------------------------------------------//
sub view_static_hosts {
    append_error_log("view_static_hosts: start\n") if $DEBUG>=1;
    my %form = get_input();
    print_html_file("admin_header.html");


    append_error_log("view_static_hosts: pre-lock\n") if $DEBUG>=1;
    # Lock the entire dhcpd_master file here to ensure corruption
    # Cannot occur.
    my $lock_file = open_with_lock("> $DHCPDMASTERLOCK");
    if (!$lock_file) {
        admin_error("Could not exclusively open $DHCPDMASTERLOCK lock file.");
        print_html_file("admin_footer.html");
    }
    append_error_log("view_static_hosts: post-lock\n") if $DEBUG>=1;

    print_static_page($form{'section'});
    close $lock_file;
    append_error_log("view_static_hosts: end\n") if $DEBUG>=1;

    print_html_file("admin_footer.html");
} 

#----------------------------------------------------------------------------\\
# force_update: Starts execution of an external script that synchronises
# between multiple remote dhcpd servers.
#
# Parameters: None.
#
# Returns: Nothing.
#----------------------------------------------------------------------------//
sub force_update {
    print_html_file("admin_header.html");
    if (!priv('force_update')) {
        admin_error("Superuser rights are required for this action. ".
                    "Sync will occur every 15 minutes automatically.",
                    "Privilege");
        print_html_file("admin_footer.html");
    }

    # Flush the output buffer after each print
    select((select(STDOUT), $|=1)[0]);

    print "Synchronisation started...<br />";

    my $output = `$SYNC_SCRIPT`;
    print $output =~ s/\n/<br>/g;
    print "Synchronisation completed";
    print_html_file("admin_footer.html");
}

#----------------------------------------------------------------------------\\
# do_logon: attempts to authenticate a user with the system, using either
# a provided username/password or a session cookie.
#
# Parameters: None.
#
# Returns: Nothing useful. Will display a logon page on failure and will exit.
#----------------------------------------------------------------------------//
sub do_logon {
    my %cookies = get_cookies();
    my %form = get_input();
    my $user;
    if ($cookies{'netreg-session'}) {
        $user  = is_valid_cookie($cookies{'netreg-session'});
    } 

    if (!$user) {
        if ($form{'user'} && $form{'pass'}) {
            # will not return if it failed.
            logon ($form{'user'}, $form{'pass'}, "1");
        } else {
            my $image = "background-image: url(\"/gfx/logo_6.jpg\");\n".
                        "background-repeat: no-repeat;\n".
                        "background-position: center center;\n";
            print_html_file('admin_header.html', "dummy;", $image);
            print_html_file('admin_logon.html');
            print_html_file('admin_footer.html');
        }
    }

    if (!priv('logon_admin')) {
        print_html_file("admin_header.html");
        admin_error("Insufficient Privileges", "Privilege");
        logoff(user());
        print_html_file("admin_footer.html");
    }
}

