#!/usr/bin/perl -w
#----------------------------------------------------------------------------\
# NetReg 2.0 (pre-release devel version)
#       DHCPD Registration Frontend
#
# Version 1.3: 
# Copyright (C) 1999-2002 Peter Valian, Southwestern University
# Version 2.0:
# Portions Copyright (C) 1999-2002 Peter Valian, Southwestern University.
# Portions Copyright (C) 2002 David Bussenschutt, Griffith University, Australia.
# Portions Copyright (C) 2002 Robert Lowe,  Lawrence University, Wisconsin.
# Portions Copyright (C) 2002-2003 Michael McMaster, Griffith University, Australia.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Version 1.3: http://www.netreg.org/
# Version 2.0: http://www.sourceforge.net/projects/netreg/
#
# Version 1.3 code contributions by:
#   Peter Valian <valianp@southwestern.edu>,  Southwestern University.
#     Author of Version 1.3rc2 
#   Bobby Clark <bobby.clark@eku.edu>, Eastern Kentucky University 
#     IMAP Authentication
#   Michael Grinnell <grinnell@american.edu>, American University
#     LDAP Authentication
#
# Version 2.0 code contributions by:
#   Robert Lowe <robert.h.lowe@lawrence.edu>, Lawrence University, Wisconsin.
#     CIDR (Subnet Masks etc) Handling/Notation
#     Alpha Testing, and bug chasing in David's Code.
#   David Bussenschutt <D.Bussenschutt@mailbox.gu.edu.au>, Griffith University, Queensland, Australia.
#     Much re-arrangement, and fixes, including but not limited to:
#     Perl 5 Syntax, and modules
#     File Locking
#     OMAPI Support
#     Remote dhcpdmanager daemon, and client support (functionally similar to omapi, but more fully implemented)
#     Improved Error Handling
#     Use of -w for improved code
#     LDAP group handling, for Permissions.
#     install_script
#   Michael McMaster <m4k3r@users.sourceforge.net>, Griffith University,  Queensland, Australia.
#     DBI/Oracle Support
#     HTML extraction/templating
#     Permissions handling (access controls)
#     Cookie based authentication
#     Interface for editing dhcp options/groups and hosts.
#     Synchronisation between hosts
#     Scalability improvements
#     Various code clean-ups
#     Consistency checker for host information 
#----------------------------------------------------------------------------/
use strict;
use warnings;
use Socket;

use Database qw (db_connect);
use Admin;
use Authenticate;
use Files_Remote;
use Subnets qw (what_is_my_registered_subnet);
use NetregMisc qw (open_with_lock);
use StaticHosts qw (get_groups parse_conf write_conf);
use Variables qw ($NSUPDATE $Key $KeyName);

# check_consistency.pl
#
# Checks for consistency between the various places data is stored for each host
# Prints the errors found to stdout and lets the user deal with them manually.

my $dbh = db_connect();
die "Could not connect to database\n" unless $dbh;

my $sql_all_hosts = $dbh->prepare("select * from interface") || die "Database failure\n";
my $sql_get_host = $dbh->prepare ("select * from interface where mac = ?") || die "Database failure\n";
my $sql_domains = $dbh->prepare ("select * from subnet_domain where domain = ?"); 
my $sql_update_subnet = $dbh->prepare ("update interface set cidr_subnet = ?,cidr_mask=? where mac = ?");
my $sql_remove_ip = $dbh->prepare ("update interface set ip = '',cidr_subnet= '',cidr_mask='' where mac = ?");
my $sql_group = $dbh->prepare("update interface set group_id = ? where mac = ?");
my $sql_date = $dbh->prepare("update interface set statusdate = ? where mac = ?");


my $static_hosts = parse_conf();
die "Could not parse dhcpd conf info\n" unless $static_hosts;

check_hosts();

$dbh->disconnect();

#################################################################
# Check all hosts
##################################################################
sub check_hosts {
    $sql_all_hosts->execute();
    my @groups = get_groups();

    while (my @row = $sql_all_hosts->fetchrow_array) {
        # check MAC is valid.
        if ($row[0] !~ /^[0-9a-f\:]+$/i) {
            print "MAC $row[0] is an invalid format (interface DB table)\n";
        }

        # Check domain is valid.
        $sql_domains->execute($row[4]);
        my @domain_info = $sql_domains->fetchrow_array();
        $sql_domains->finish();
        if (!$domain_info[0]) {
            print "MAC $row[0] has an invalid domain\n";
        }

        # Check that date is valid
        # Dates that are too large will crash the web interface.
        if ($row[6] !~ /(\d\d\d\d)(\d\d)(\d\d)/) {
            print "MAC $row[0] has an invalid date. This may crash the web interface\n";
        }
        my $year = $1 > 2030 ? 2030 : $1;
        my $month = $2 > 12 ? 12 : $2;
        my $day = $3 > 31 ? 31 : $3; 
        $sql_date->execute("$year$month$day",$row[0]);
        $sql_date->finish();

        if ($row[5] eq 'S') {
            my ($subnet, $mask) = what_is_my_registered_subnet($row[9]);
            if (!($subnet && $mask)) {
                print "Could not find subnet for static host $row[0]!\n";
            } else {
                $sql_update_subnet->execute($subnet,$mask,$row[0]);
                $sql_update_subnet->finish();
            }

            if (!$row[9]) {
                print "No IP specified for static host $row[0]\n";
            } else {
                my $dns_ip = gethostbyname("$row[1].$row[4]");
                if ($dns_ip && inet_ntoa($dns_ip) ne $row[9]) {
                    print "DNS ip address for $row[0] conflicts with db\n";
                } elsif (!$dns_ip) {
                    print "DNS A record for $row[1].$row[4] not present. Adding...\n";
                    open (NSUPDATE, "| $NSUPDATE -y $KeyName:$Key");
                    print NSUPDATE "update add $row[1].$row[4] 3600 A $row[9]\n\n";
                    close NSUPDATE;
                } 
       
                my $dns_host = gethostbyaddr(inet_aton($row[9]), AF_INET);
                if ($dns_host && $dns_host ne "$row[1].$row[4]") {
                    print "DNS reverse mapping incorrect for $row[0]. Fixing...\n";
                    $row[9] =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/;
                    my $reverse = "$4.$3.$2.$1.IN-ADDR.ARPA";
                    open (NSUPDATE, "| $NSUPDATE -y $KeyName:$Key");
                    print NSUPDATE "update delete $reverse\n\n";
                    print NSUPDATE "update add $reverse 3600 PTR $row[1].$row[4]\n\n";
                    close NSUPDATE;

                } elsif (!$dns_host) {
                    print "DNS reverse mapping not present for $row[1].$row[4]. Adding...\n"; 
                    $row[9] =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/;
                    my $reverse = "$4.$3.$2.$1.IN-ADDR.ARPA";
                    open (NSUPDATE, "| $NSUPDATE -y $KeyName:$Key");
                    print NSUPDATE "update add $reverse 3600 PTR $row[1].$row[4]\n\n";
                    close NSUPDATE;
                }
            }
            
        } else {
            # Dynamic host only checks
           
            # Destroy any IP/subnet info saved in the DB, as this is irrelevant. 
            $sql_remove_ip->execute($row[0]);
            $sql_remove_ip->finish();
        }

        # Check the group_id
        my $found_group = 0;
        foreach (@groups) {
            if ($_ eq $row[12]) {
                $found_group++;
                last;
            }
        }
        if (!$found_group) {
            my $non_group = $row[12] || 'Unknown';
            print "Host $row[0] is in a non-existent group $non_group\n";
            $sql_group->execute('Roaming:Default',$row[0]);
            $sql_group->finish();
            print "Fixed. Group changed to Roaming:Default.\n";
        }
    }
    $sql_all_hosts->finish();
}
